import numpy as np
from .ClusterCrossLinking import CrossLinkIndex
from .times_2_title import *
from SpringSaLaDpy.data_locator import *
from .Format import format

def plot(search_directory, times, size_threshold_mean=1, grouping='all', hist=False):     
    input_file, rounded_times, title_str = format(search_directory, times, 'viewer')

    vf = data_file_finder(search_directory, ['viewer_files'], run=0)

    CLI = CrossLinkIndex(input_file, ss_timeSeries=rounded_times)

    print(CLI)
    #d = cl.mapSiteToMolecule()
    #rif = ReadInputFile(txtfile)
    #print(rif.getReactiveSites())
    #print(len(cl.getActiveSiteIDs())) 
    CLI.getSI(vf) 
    CLI.getSI_stat() 
    CLI.plot_SI_stat(color='gray', fs=16, xticks=None, yticks=None, hist=hist, title_str=title_str, size_threshold_mean=size_threshold_mean, grouping=grouping)
    #CLI.plot_SI_stat(color='c', xticks=None, yticks=None)


import os
import fnmatch
import glob

#Finds the last txt file in the search_directory
def find_txt_file(search_directory, search_term=''):
    if search_term != '':
        return find_files(search_directory, search_term)
    else:
        match = None
        for file in os.listdir(search_directory):
            if file.endswith(".txt"):
                match = os.path.join(search_directory, file)
        return match

#Finds the last file in a given directory that contains search_term in the file name
def find_files(search_directory, search_term, no_warnings=False):
    match = None
    count = 0
    for filename in os.listdir(search_directory):
        if search_term in filename:
            count += 1
            match = os.path.join(search_directory, filename)
    if not no_warnings:
        if match == None:
            print(f'Error, could not find the search term "{search_term}" in any file in the search directory "{search_directory}"')
        if count > 1:
            print(f'Warning, there are multiple files that contain the search term "{search_term}".\nThe file "{match}" is currently selected.\nPlease provide a more specific search term if this is not the file you want\n')
    return match

def data_file_finder(search_directory, path_list, search_term = None, file_name = None, run = None, no_warnings=True):
    if file_name == None:
        file_name = find_txt_file(search_directory)
    else:
        file_name = os.path.join(search_directory, file_name)
    
    augmented_path_list = []
    if search_directory[-7:] == '_FOLDER':
        augmented_path_list = [search_directory] + path_list
    else:
        augmented_path_list = [file_name[:-4] + '_FOLDER'] + path_list
    lower_search_directory = os.path.join(*augmented_path_list)

    if run != None:
        result = find_files(lower_search_directory, f'Run{run}', no_warnings=no_warnings)
    else:
        result = find_files(lower_search_directory, search_term, no_warnings=no_warnings)
    return result


from .MoleculeCounter import *
from .Molclustpy_visualization_funcitons import *
from SpringSaLaDpy.display_info import *
from SpringSaLaDpy.data_locator import *

def plot(search_directory, data_selection='FullCount', indices=[], list_options=True, legend_right=True, fill=True):
    data_file_path = data_file_finder(search_directory, search_term=data_selection, path_list=['data', 'Run0'])
    data_file = os.path.split(data_file_path)[1]
    
    # Compute the free molecular concentrations
    mc = MoleculeCounter(search_directory, dataFile=data_file)
    mc.getMoleculeStat()

    title = os.path.split(data_file)[1][:-4]
    if list_options:
        column_info(data_file_path, start_col=1)

    plotTimeCourseCopy(search_directory, title, indices, legend_right, fill)


import ast
import pandas as pd
import numpy as np
from .Molclustpy_visualization_funcitons import *
from SpringSaLaDpy.data_locator import *

def composition_calc(search_directory, title_str, special_clusters, width=0.1, alpha=0.6, legend_right=True):

    path = data_file_finder(search_directory, ['pyStat', 'Cluster_stat'], search_term='Clusters_composition')

    with open(path) as file:
        lines = file.readlines()
    file.close()

    molecules_list = ast.literal_eval(lines[0].split('\t')[1][1:-13])
    columns = ['Clusters']
    columns.extend(molecules_list)

    output_lists = []
    for line in lines:
        if line != '\n' and line[:12] != 'Cluster Size':
            
            size_list = line.split('\t')
            size_list.pop()
            size_list.pop(0)
            size_list.pop(0)

            for i, item in enumerate(size_list):
                size_list[i] = size_list[i].split(' : ')
                tmp = size_list[i][0].split(',')
                size_list[i][0] = np.array([float(entry) for entry in tmp])*float(size_list[i][1][:-1])/100
                size_list[i] = size_list[i][0]
            
            summed_list = sum(size_list)
            size = round(sum(summed_list))
            relative_list = summed_list/sum(summed_list)

            output_list = [size]
            output_list.extend(relative_list)
            output_lists.append(output_list)
    df = pd.DataFrame(output_lists, columns=columns)

    outpath = os.path.join(os.path.split(path)[0], 'Cluster_composition.csv')
    
    df.to_csv(outpath)

    plotClusterCompositionCopy(outpath, title_str, specialClusters=special_clusters, width=width, alpha=alpha, legend_right=legend_right)


from SpringSaLaDpy.input_file_extraction import *
import os
import matplotlib.pyplot as plt
from .ClusTopology_ss import ClusterDensity
from .DataPy import *
from .times_2_title import * 
from .Format import format

def read_viewer(path):
    if path[-7:] == '_FOLDER':
        last_item = ''
    else:
        last_item = os.path.split(path)[1][:-12] + '_FOLDER'
    specific_path  = os.path.join(path, last_item)

    _,split_file = read_input_file(specific_path)
    total_time = float(split_file[0][1].split(' ')[2])
    dt_image = float(split_file[0][5].split(' ')[1])
    count = int(total_time/dt_image)

    input_file = find_txt_file(specific_path)
    return count, dt_image, input_file

def scatter_plot(search_directory, times=[], size_threshold=1, cubic_com=True):
    
    input_file, rounded_times, title_str = format(search_directory, times, file_type='viewer')

    cd = ClusterDensity(input_file, ss_timeSeries=rounded_times)
    cd.getCD_stat(cs_thresh=size_threshold, title_str=title_str, cubic_com=cubic_com)

def histogram(search_directory, times=[], size_threshold=1, indices=[], cubic_com=True):
    
    input_file, rounded_times, title_str = format(search_directory, times, file_type='viewer')

    cd = ClusterDensity(input_file, ss_timeSeries=rounded_times)
    cd.getCD_stat(cs_thresh=size_threshold, title_str=title_str, hist=True, indices=indices, cubic_com=cubic_com)

def time_course(path, indices = [], size_threshold=1, legend_right=True, cubic_com=True, reference_shapes=False):
    count, dt_image, input_file = read_viewer(path)

    times = [0]
    for i in range(int(count) + 1):
        times.append(i*dt_image)
    times.pop(0)

    cd = ClusterDensity(input_file, ss_timeSeries=[times[0]])
    vfiles = glob(cd.simObj.getInpath() + "/viewer_files/*.txt")[:]
    
    outputs = []
    for i in range(4):
        outputs.append([[],[],[]])

    for time in times:
        cs_tmp, rg_tmp, rmax_tmp = [], [], []
        
        cd = ClusterDensity(input_file, ss_timeSeries=[time])
        for vfile in vfiles:
            #this line takes up the majority of the computation time
            res, MCL, mtp_cs, mtp_rg, mtp_rmax = cd.getClusterDensity(vfile, size_threshold, cubic_com=cubic_com)

            cs_tmp.extend(mtp_cs)
            rg_tmp.extend(mtp_rg)
            rmax_tmp.extend(mtp_rmax)

        csList = np.concatenate(cs_tmp).ravel().tolist()
        rgList = np.concatenate(rg_tmp).ravel().tolist()
        rmaxList = np.concatenate(rmax_tmp).ravel().tolist()
        distribution = [(rg/rmax) for rg, rmax in zip(rgList, rmaxList)]
        
        plotting_lists = [rgList, csList, rmaxList, distribution]

        for i, plotting_list in enumerate(plotting_lists):
            if csList == []:
                for j in range(3):
                    outputs[i][j].append(None)
            else:
                outputs[i][0].append(min(plotting_list))
                outputs[i][1].append(sum(plotting_list)/len(plotting_list))
                outputs[i][2].append(max(plotting_list))

    plot_dict = {
        0: 'Minimum',
        1: 'Average',
        2: 'Maximum'
    }

    label_dict = {
        0: ['Radius of Gyration', 'Radius of Gyration (nm)'],
        1: ['Cluster Size', 'Molecules per Cluster'],
        2: ['Cluster Radius', 'Radius (nm)'],
        3: ['Mass Distribution Coefficient', 'Radius of Gyration / Radius']
    }

    if indices == []:
        indices = [0,1,2]

    legend_list = []
    for i, output in enumerate(outputs):
        plt.figure()
        lines = []
        for index in indices:
            legend_list.append(plot_dict[index])
            line, = plt.plot(times,output[index])
            lines.append(line)
            plt.title(label_dict[i][0])
            plt.ylabel(label_dict[i][1])
        plt.xlabel('Time (seconds)')
        
        if legend_right:
            top_legend = plt.legend(lines, legend_list, bbox_to_anchor=(1.02, 1), loc='upper left')
        else:
            top_legend = plt.legend(lines, legend_list)      

        if i == 3 and reference_shapes:
            plt.gca().add_artist(top_legend)
            shell = plt.axhline(1, ls='-', lw=1.5, color='k', label='Thin Spherical Shell = 1')
            sphere = plt.axhline(np.sqrt(3/5), ls='dashed', lw=1.5, color='k', label=r'Uniform Solid Sphere = $\sqrt{\frac{3}{5}}$')
            if legend_right:
                plt.legend(handles=[shell, sphere], bbox_to_anchor=(1.02, 0), loc='lower left')
            else:
                plt.legend(handles=[shell, sphere], loc='lower right')

def scatter_3d(path, run=0, times=[], view=[], site_scale=10, link_thickness=0.75, depth_fade=True):  
    count, dt_image, input_file = read_viewer(path)
    if times == []:
        times.append(-1)
    for time in times:
        if time == -1:
            time_input = []
        else:
            time_input = [time]
        _, rounded_times, title_str = format(path, time_input, 'cluster')
        cd = ClusterDensity(input_file, ss_timeSeries=[rounded_times[0]])
        vfiles = glob(cd.simObj.getInpath() + "/viewer_files/*.txt")[:]

        viewerfile = vfiles[run]

        with open(viewerfile, 'r') as infile:
            tps, index_pairs = cd.getSteadyStateFrameIndices(viewerfile)
            lines = infile.readlines()
            for i,j in index_pairs:
                #i,j = index_pairs[-1]
                current_frame = lines[i:j]
                posDict, linkList, colorDict, radDict = cd.getBindingStatus(current_frame)

        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(111, projection='3d')

        x_cords=[]
        y_cords=[]
        z_cords=[]
        colors=[]
        radii = []
        for site in posDict.keys():
            x_cords.append(posDict[site][0])
            y_cords.append(posDict[site][1])
            z_cords.append(posDict[site][2])
            colors.append(colorDict[site])
            radii.append(site_scale*radDict[site])

        connection_list = []
        for link in linkList:
            x_pair = [posDict[link[0]][0], posDict[link[1]][0]]
            y_pair = [posDict[link[0]][1], posDict[link[1]][1]]
            z_pair = [posDict[link[0]][2], posDict[link[1]][2]]
            connection_list.append([x_pair, y_pair, z_pair])

        if depth_fade:
            alpha = None
        else:
            alpha = 1
        ax.scatter(x_cords, y_cords, z_cords, c=colors, s=radii, alpha=alpha)
        for connection in connection_list:
            ax.plot(connection[0], connection[1], connection[2], color='black', linewidth=link_thickness)
        
        if view==[]:
            pass
        else:
            ax.view_init(view[0], view[1], view[2])

        #ax.set_aspect('equal')
        ax.set_xlabel('X (nm)')
        ax.set_ylabel('Y (nm)')
        ax.set_zlabel('Z (nm)')
        ax.set_xlim(cd.box[0][0], cd.box[0][1])
        ax.set_ylim(cd.box[1][0], cd.box[1][1])
        ax.set_zlim(cd.box[2][0], cd.box[2][1])

        ax.set_title(f'3D View of Sites and Links{title_str}')

        plt.show()
    


        
        

        
        
    

    
    


import re
from .input_file_extraction import *
import tempfile
from .Project_reader import *
from .data_locator import *
import os

#add output path options
def edit_input_file(search_directory, output_path='', counts=[], on_rate=[], off_rate=[]):

    path = find_txt_file(search_directory)
    molecules, split_file = read_input_file(search_directory)

    molecule_descriptions = []
    reaction_descriptions = []
    for molecule in molecules:
        if molecule[0][0:8] == 'MOLECULE':
            molecule_descriptions.append(molecule[0].split())

    if counts != []:
        for i, count in enumerate(counts):
            molecule_descriptions[i][4] = count
            molecule_descriptions[i] = (' '.join(map(str, molecule_descriptions[i])))

    for section in split_file:
        if section[0][:15] == '*** BIMOLECULAR':
            for reaction in section[2:-1]:
                reaction_descriptions.append(reaction.split())

    if on_rate != []:
        for i, count in enumerate(on_rate):
            diff = len(reaction_descriptions[i]) - 18
            reaction_descriptions[i][13 + diff] = count
    
    if off_rate != []:
        for i, count in enumerate(off_rate):
            diff = len(reaction_descriptions[i]) - 18
            reaction_descriptions[i][15 + diff] = count

    for i in range(len(reaction_descriptions)):
        reaction_descriptions[i] = (' '.join(map(str, reaction_descriptions[i])))

    with open(path, 'r') as file: 
        data = file.readlines() 
    file.close()

    if counts != []:
        count = 0
        for i, line in enumerate(data):
            if line[0:11] == 'MOLECULE: "':
                data[i] = molecule_descriptions[count] + '\n'
                count = count + 1
    
    if on_rate != [] or off_rate != []:
        for i, line in enumerate(data):
            if line[:15] == '*** BIMOLECULAR':
                for j, item in enumerate(reaction_descriptions):
                    data[i+j+2] = item + '\n'

    #file_name = "output.txt"
    #file_path = os.path.join(search_directory, file_name)

    #warn if the input file is going to be overwritten
    
    #os.makedirs(output_path, exist_ok=True)

    if output_path=='' or os.path.exists(output_path):
        if output_path=='':
            output_path=path
        ans = input(f'Warning, {output_path} is about to be overwritten, are you sure you want to continue? (Y/N)')
        if ans.upper() == 'Y':
            with open(output_path, 'w+', encoding='utf-8') as file:
                file.writelines(data)
            file.close()
        else:
            print('Edit operation canceled')
    else:
        with open(output_path, 'w+', encoding='utf-8') as file:
            file.writelines(data)
        file.close()


import os
from .data_locator import *

def read_input_file(directory_path, search_term=''):
    path = find_txt_file(directory_path, search_term)

    with open(path, "r") as f:
        lines = f.readlines()

    split_file = []
    current_list = []

    for line in lines:
        if line[0:3] == '***':
            split_file.append(current_list)
            current_list = []
        current_list.append(line.strip())
    split_file.append(current_list)
    split_file.pop(0)

    if len(split_file) < 10:
        print(f'Error: Input file {path} doesn\'t match the expected format. Please ensure your input file is unmodified and that there are no other .txt files at the top level of your search directory.')

    #Extract molecule data
    molecules = []
    molecule = []
    for line in split_file[2]:
        if line[0:9] == 'MOLECULE:':
            molecules.append(molecule)
            molecule = []
        molecule.append(line)
    molecules.append(molecule)    
    molecules.pop(0)

    return(molecules, split_file)

def read_reactions(split_file):
    #Extract reaction data
    transition_reactions = []
    allosteric_reactions = []
    binding_reactions = []
    reaction_types = [['transition_reactions', 5], ['allosteric_reactions', 6], ['binding_reactions', 7]]
    for reaction_type in reaction_types:
        for line in split_file[reaction_type[1]]:
            if line != '':
                exec(reaction_type[0] + '.append(line)')
        exec(reaction_type[0] + '.pop(0)')
    return(transition_reactions, allosteric_reactions, binding_reactions)


import pandas as pd
from SpringSaLaDpy.data_locator import data_file_finder
from SpringSaLaDpy.input_file_extraction import read_input_file
from .Molclustpy_visualization_funcitons import *
from SpringSaLaDpy.time_rounder import find_nearest_time
import os

def plot(search_directory, time=None, specialClusters=[], width=0.15, alpha=0.5):
    path_list = ['data', 'Cluster_stat', 'Histograms', 'Size_Comp_Freq', 'MEAN_Run']
    
    #Round to nearest available time based on dt_data value
    _, split_file = read_input_file(search_directory)
    dt_data = float(split_file[0][4][9:])

    search_term = 'Size_Comp_Freq'

    time, comp_file = find_nearest_time(search_directory, path_list, time, dt_data, search_term)

    df = pd.read_csv(comp_file)
    df = df.rename({'Size':'Clusters'}, axis = 1)

    molecules = df.columns[1].split(',')

    df_expanded = df[df.columns[1]].str.split(',', expand=True)
    df_expanded.columns = molecules

    df = pd.concat([df, df_expanded], axis=1)

    for molecule in molecules:
        df[molecule] = df[molecule].astype(float) * df['Frequency in clusters of the same size']

    df = df.groupby('Clusters')[molecules].sum().reset_index()

    df['row_sum'] = df[molecules].sum(axis=1)

    for molecule in molecules:
        df[molecule] = df[molecule] / df['row_sum']

    df = df.drop(['row_sum'], axis=1)

    outpath = os.path.normpath(comp_file)
    outpath = os.path.join(*outpath.split(os.sep)[:-5])

    if not os.path.isdir(outpath + os.sep + 'pyStat'):
        os.makedirs(outpath + os.sep + 'pyStat')

    csv_path = os.path.join(outpath, 'pyStat', 'Cluster_composition.csv')
    df.to_csv(csv_path)

    plotClusterCompositionCopy(outpath, time, specialClusters, width, alpha)


from .ssalad_ClusterAnalysis import *
from .Molclustpy_visualization_funcitons import *
from .Format import format

def plot(search_directory, times, bins=[]):
    
    input_file, rounded_times, title_str = format(search_directory, times, 'cluster')

    ca = ClusterAnalysis(input_file)
    ca.getMeanTrajectory(SingleTraj=False)
    ca.getSteadyStateDistribution(SS_timePoints=rounded_times)

    plotClusterDistCopy(search_directory, rounded_times, bins, title_str)


import numpy as np

def location2vec(location_name):
    match location_name:
        case 'Intracellular':
            return np.array([1, 0, 0])
        case 'Extracellular':
            return np.array([0, 1, 0])
        case 'Membrane':
            return np.array([0, 0, 1])


import os
from .data_locator import data_file_finder
from .digit_counter import count_digits

def find_nearest_time(search_directory, path_list, time, dt, search_term):
    if time != None:
        #Round to nearest available time based on dt value
        if round(time % dt, 7) >= dt/2:
            time = round(time - (time % dt) + dt, 7)
        else:
            time = round(time - (time % dt), 7)

        decimals = count_digits(dt, 7)
        time = format(float(time), f'.{decimals}f')
        file = data_file_finder(search_directory, path_list, search_term=time, no_warnings=True)
    else:
        file = data_file_finder(search_directory, path_list, search_term=search_term, no_warnings=True)
        time = float(os.path.split(file)[1].split('_')[2])
    return time, file




import os
import glob
import pandas as pd

def convert_bytes(num):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0

#Displays all the files in the same directory as the input file (given by a path)
def file_info(path, list_neighbors=False):
    print('Selected File:\n' + os.path.split(path)[1] + '\n')
    print('Parent Directory:\n' + os.path.split(path)[0] + '\n')
    search_pattern = os.path.join(os.path.split(path)[0], '*Average*')
    entries = glob.glob(search_pattern)
    
    if list_neighbors:
        print('Average Files:')
        for entry in entries:
            #print(os.path.split(entry)[1][23:], )
            print('{0:50}  {1}'.format(os.path.split(entry)[1][23:], convert_bytes(os.path.getsize(entry))))

def column_info(path=None, frame=None, start_col=0):
    if frame is None:
        df = pd.read_csv(path)
        entries = int((df.shape[1] - 2))
    else:
        df = frame
        entries = int(df.shape[1])

    print('Columns:')
    for i in range(start_col,entries+1):
        print(f'{i-start_col}: {df.columns[i].strip()}')

    lines = []
    for i in range(start_col,entries + 1):
        lines.append(i-start_col)

    print('\nList of indices:')
    print(lines)


from decimal import Decimal, getcontext

def count_digits(number, precision):
    getcontext().prec = precision
    decimal_number = Decimal(str(number))
    decimal_places = -decimal_number.as_tuple().exponent

    return decimal_places


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 25 16:27:54 2019

@author: Ani Chattaraj
"""
import re, json 
import os, sys
from decimal import Decimal
import numpy as np
import pandas as pd
from csv import writer
from collections import defaultdict, OrderedDict
from time import time
from glob import glob
#from DataPyPlot import DataPyPlot
import matplotlib.pyplot as plt
import multiprocessing as mp

def ProgressBar(jobName, progress, length=40):
    completionIndex = round(progress*length)
    msg = "\r{} : [{}] {}%".format(jobName, "*"*completionIndex + "-"*(length-completionIndex), round(progress*100))
    if progress >= 1: msg += "\r\n"
    sys.stdout.write(msg)
    sys.stdout.flush()

def displayExecutionTime(func):
    """
    This decorator will calculate time needed to execute a function 
    """
    def wrapper(*args, **kwrgs):
        t1 = time()
        func(*args, **kwrgs)
        t2 = time()
        delta = t2 - t1
        if delta < 60:
            print("Execution time : {:.4f} secs".format(delta))
        else:
            t_min, t_sec = int(delta/60), round(delta%60)
            print(f"Execution time : {t_min} mins {t_sec} secs")
    return wrapper

def getSteadyStateIndices(timeSeries, ss_timepoints, rel_tol=1e-8):
    # both are lists of float numbers; so cant' apply equally directly
    ss_index = []
    for i, elem in enumerate(timeSeries):
        if any([np.isclose(elem, t, rtol=rel_tol) for t in ss_timepoints]):
            # returns true if finds atleast one match
            ss_index.append(i)
    return ss_index

def getCompleteRunList(txtfile):
    rif = ReadInputFile(txtfile)
    path = rif.getInpath() + "/pyStat"
    with open(path + "/LCR.txt", 'r') as tmpfile:
        lcr = tmpfile.readlines()[0]
        LCR = lcr.split(',')[:-1]
        LCR = [int(r) for r in LCR]
    return LCR

    

class ReadInputFile:
    
    def __init__(self, txtfile):
        self.txtfile = txtfile
    
    def readFile(self):
        with open(self.txtfile, 'r') as tmpfile:
            txtfile = [line for line in tmpfile.readlines()]
        return txtfile
    
    def getInpath(self):
        mypath = self.txtfile.replace('\\','/').split("/")[:-1]
        return "/".join(mypath)
    
    def getOutpath(self, statName):
        # statName : Name of the analysis; cluster_stat, e2e_stat etc
        inpath = self.getInpath()
        outpath = inpath + f"/pyStat/{statName}"
        if not os.path.isdir(outpath):
            os.makedirs(outpath)
        return outpath
    
    def getMolecules(self):
        lines = self.readFile()
        molNames, molCounts = [], []
        
        for index, line in enumerate(lines):
            if not re.search('[*]', line):
                if re.search('MOLECULE', line):
                    specs = line.split(':')[-1].split()
                    if len(specs)>2:
                        if int(specs[3]) != 0:  # molecules with non-zero count
                            molNames.append(specs[0].replace('"',''))
                            molCounts.append(int(specs[3]))
                            
        return molNames, molCounts
    
    @staticmethod
    def StringSearch(string, lines):
        value = None
        for line in lines:
            if re.search(string, line):
                value = line.split(':')[-1]
                break
        return value
    
    def getTimeStats(self):
        lines = self.readFile()
        t_tot = self.StringSearch("Total time", lines)
        ts = self.StringSearch("dt", lines)  # ts : time step for simulation
        t_data = self.StringSearch("dt_data", lines)
        t_image = self.StringSearch("dt_image", lines)
        return float(t_tot), float(ts), float(t_data), float(t_image)
    
    def getBoundingBox(self):
        lines = self.readFile()
        x_bound = self.StringSearch("L_x", lines)
        y_bound = self.StringSearch("L_y", lines)
        z_bound_low = self.StringSearch("L_z_out", lines)
        z_bound_high = self.StringSearch("L_z_in", lines)
        return [(-1e3*float(x_bound)/2, 1e3*float(x_bound)/2), (-1e3*float(y_bound)/2, 1e3*float(y_bound)/2), (-1e3*float(z_bound_low), 1e3*float(z_bound_high))]
    
    def getNumRuns(self):
        numRuns = 0
        lines = self.readFile()
        for line in lines:
            if re.search("\\b" + "Runs" + "\\b", line): # \\b : word boundary; avoids the line 'SimultaneousRuns: 1'
                numRuns = line.split(':')[-1]
                break
        #numRuns = len(os.listdir(os.path.join(self.getInpath(), 'viewer_files')))
        return int(numRuns)
    
    def getReactiveSites(self):
        reactive_sites = []
        lines = self.readFile()
        n1, n2 = 0, 0
        for index, line in enumerate(lines):
            if re.search("BINDING REACTIONS", line):
                n1 = index
            elif re.search("MOLECULE COUNTERS", line):
                n2 = index
        binding_rxns = lines[n1+1 : n2]
        for line in binding_rxns:
            if line != '\n':
                lhs, rhs = [elem for elem in line.split('+')]
                lhs = lhs.split(':')
                rhs = rhs.split(':')
                reactive_sites.append(lhs[-2])
                reactive_sites.append(rhs[1])
        
        reactive_sites = [site.replace("'","").strip() for site in reactive_sites]
        reactive_sites = [site for site in set(reactive_sites)]
        
        return reactive_sites
    

class ClusterAnalysis:
    
    def __init__(self, simfile, t_total=None, runs=None, withMonomer=True):
        
        """
        simfile: string 
        t_total: float;
        """
        self.simfileObj = ReadInputFile(simfile)
        
        tf, dt, dt_data, dt_image = self.simfileObj.getTimeStats()
        self.dt = dt_data
        if t_total == None:
            self.t_total = tf
        else:
            self.t_total = t_total
        self.withMonomer = withMonomer
        
        timepoints = np.arange(0, self.t_total+self.dt, self.dt)
        self.timePoints = timepoints
        
        numRuns = self.simfileObj.getNumRuns()
        if runs == None:
            self.runs = [run for run in range(numRuns)]
        elif type(runs) is int:
            self.runs = [i for i in range(runs)]
        elif type(runs) is list:
            self.runs = runs
        else:
            print("runs must be an integer or a list of runs")
            
    def __repr__(self):
        simfile = self.simfileObj.txtfile.split('/')[-1]
        info = f"Class : {self.__class__.__name__}\nSystem : {simfile}\nTotal_Time : {self.t_total} seconds\t\tnumRuns : {len(self.runs)}"
        return info
    
    @staticmethod        
    def getSpeciesCount(df, species):
        # df : pandas dataframe
        col1, col2 = df[0], df[1]
        return [col2[i] for i in range(len(col1)) if col1[i]== species]
    
    def getClusterAverage(self, csvfile):
        # including the monomer
        acs, aco, foTM = 0, 0, 0
        clusterList = []
        df = pd.read_csv(csvfile, header=None)
        clusterCount = df[1][0]
        #col1, col2 = df[0], df[1]
        if clusterCount == 0:
            pass
            #print('\nNo cluster found')
        else:
            oligos = self.getSpeciesCount(df, "Size")
            if len(oligos) == 0:
                acs, aco = 1, 1
            else:
                monomerCount = clusterCount - len(oligos)
                clusterList.extend(oligos)
                clusterList.extend([1]*monomerCount)

                N, TM = len(clusterList), sum(clusterList)
                acs = TM/N # acs : average cluster size
                foTM = {clus: (clusterList.count(clus)*(clus/TM)) for clus in set(clusterList)} #foTM : fraction of total molecules
                aco = sum([cs*f for cs, f in foTM.items()])  # aco : average cluster occupancy

        #print(csvfile)
        #print('foTM', foTM)
        return acs, aco
    
    def getMonomerCounts(self, csvfile, mols=[], counts=[]):
        df = pd.read_csv(csvfile, header=None)
        mols_in_clus = [sum(self.getSpeciesCount(df, mol)) for mol in mols]
        monomers = np.array(counts) - np.array(mols_in_clus)
        return list(monomers)
    
    
    def getOligoAverage(self, csvfile):
        # excluding the monomer
        acs, aco = 1, 1
        oligoList = []
        df = pd.read_csv(csvfile, header=None)
        oligos = self.getSpeciesCount(df, "Size")
        N, TMC = len(oligos), sum(oligos)  # TMC: total molecules in cluster
        try:
            acs = TMC/N
            foTM = {oligo: (oligos.count(oligo)*(oligo/TMC)) for oligo in set(oligos)}
            aco = sum([cs*f for cs, f in foTM.items()])
        except:
            pass # if there is no cluster > 1, acs = aco = 1
        return acs, aco
    
    ''' 
    @staticmethod
    def getTimeAverage(ndList, numRuns, numTimePoints):
        ndArray = np.array(ndList).reshape(numRuns, numTimePoints)
        timeAverage = np.mean(ndArray, axis=0, dtype=np.float64) #float64 gives more accurate mean
        return timeAverage
    
    @displayExecutionTime
    def getMeanTrajectory(self):
        print("Getting Clustering Timecourse ...")
        simObj = self.simfileObj
        inpath = simObj.getInpath()
        if not self.withMonomer:
            outpath = simObj.getOutpath('Cluster_stat_wom')
        else:
            outpath = simObj.getOutpath("Cluster_stat")
        
        nf = abs(Decimal(str(self.dt)).as_tuple().exponent) # Number of decimal points to format the clusterTime.csv file
        acs_stat, aco_stat = [], []
        timepoints = np.arange(0, self.t_total+self.dt, self.dt)
        numRuns = len(self.runs)
        for i, run in enumerate(self.runs):
            for tp in timepoints:
                df = pd.read_csv((inpath + "/data/Run{}/Clusters_Time_{:.{}f}.csv".format(run, tp, nf)), header=None)
                if self.withMonomer:
                    acs, aco = self.getClusterAverage(df, tp, run) # need to change this func
                else:
                    acs, aco = self.getOligoAverage(df)
                
                acs_stat.append(acs)
                aco_stat.append(aco)
            ProgressBar("Progress", (i+1)/numRuns)
        
        mean_acs = self.getTimeAverage(acs_stat, numRuns, len(timepoints))
        mean_aco = self.getTimeAverage(aco_stat, numRuns, len(timepoints))
        
        tp_ms = timepoints*1e3 # second to millisecond
        
        with open(outpath+"/Clustering_dynamics.csv","w", newline='') as outfile:
            wf = writer(outfile,delimiter=',')  # csv writer
            wf.writerow(['Time (ms)','ACS','ACO'])
            wf.writerows(zip(tp_ms,mean_acs,mean_aco))
            outfile.close()
    '''
    
    @displayExecutionTime
    def getMeanTrajectory_modified(self, saveClusDist=False, process='SEQ', saveMultiTraj=False):
        # process : 'seq' (sequential) or 'par' (parallel)
        # saveClusdist = saves timecourse of cluster distribution as a function of time
        print("Getting Clustering Timecourse by modified method ...")
        simObj = self.simfileObj
        molecules, counts = simObj.getMolecules()
        inpath = simObj.getInpath()
        if not self.withMonomer:
            outpath = simObj.getOutpath('Cluster_stat_wom')
        else:
            outpath = simObj.getOutpath("Cluster_stat")

        nf = abs(Decimal(str(self.dt)).as_tuple().exponent) # Number of decimal points to format the clusterTime.csv file

        timepoints = self.timePoints
        numRuns = len(self.runs)
        numTimePoints = len(timepoints)
        csvfiles = [inpath + f"/data/Run{run}/Clusters_Time_{tp:.{nf}f}.csv" for run in self.runs for tp in timepoints]
        N = len(csvfiles)

        if process.upper() == 'SEQ':
            print('Sequential processing...')
            #ave_clus_stat = list(map(self.getClusterAverage, csvfiles))
            if not self.withMonomer:
                ave_clus_stat = [*map(lambda x: [self.getOligoAverage(x), ProgressBar('Oligo_ave_calc', progress=(csvfiles.index(x)+1)/N)] , csvfiles)]
            else:
                ave_clus_stat = [*map(lambda x: [self.getClusterAverage(x), self.getMonomerCounts(x,molecules,counts), ProgressBar('Cluster_ave_calc', progress=(csvfiles.index(x)+1)/N)] , csvfiles)]
                
            ave_stat, monomers, _ = list(zip(*ave_clus_stat)) # acs_aco, None
            acs, aco = list(zip(*ave_stat))
            

        elif process.upper() == 'PAR':
            print('Parallel processing...')
            pool = mp.Pool(4)
            #ave_clus_stat = pool.starmap(lambda x: [self.getClusterAverage(x), ProgressBar('Cluster_ave_calc', progress=(csvfiles.index(x)+1)/N)] , csvfiles)
            if not self.withMonomer:
                ave_clus_stat = pool.map(self.getOligoAverage, csvfiles)
            else:
                ave_clus_stat = pool.map(self.getClusterAverage, csvfiles)
            
            acs, aco = list(zip(*ave_clus_stat))

        print('Done')

        
        #print(foTM)
        mean_acs =np.mean(np.array(acs).reshape((numRuns, numTimePoints)), axis=0)
        mean_aco =np.mean(np.array(aco).reshape((numRuns, numTimePoints)), axis=0)
        stdev_aco =np.std(np.array(aco).reshape((numRuns, numTimePoints)), axis=0)
        #mean_monomers = np.mean(np.array(monomers).reshape((numRuns, numTimePoints)), axis=0)
        
        
        mono_traj = np.split(np.array(monomers), numRuns)
        m_traj = np.mean(mono_traj, axis=0)
        #print(m_traj)
        tp_ms = timepoints*1e3 # second to millisecond
        
        if saveMultiTraj:
            aco_trajs = np.array(aco).reshape((numRuns, numTimePoints)).T
            
            run_counts = [f'Run_{run}' for run in range(numRuns)]
            header = '\t'.join(run_counts)
            
            with open(outpath + "/MultiTraj_stat.txt", "w") as tmf00:
                np.savetxt(tmf00, aco_trajs, fmt='%.3e', header=header)
        
        with open(outpath + "/Monomer_counts.txt", "w") as tmf01:
            head_string =  '\t'.join(molecules)
            #tp_tmp = [[t] for t in tp_ms]
            #np.insert(m_traj, [:0], tp_ms, axis=1)
            np.savetxt(tmf01, m_traj, fmt='%.5e', header=head_string)
            #print(m_traj)

        with open(outpath + "/Clustering_dynamics.csv","w", newline='') as outfile:
            wf = writer(outfile,delimiter=',')  # csv writer
            wf.writerow(['Time (ms)','ACS','ACO','Stdev_ACO'])
            wf.writerows(zip(tp_ms, mean_acs, mean_aco, stdev_aco))
            outfile.close()
        
        
    @staticmethod    
    def writeComposition(outpath, compo_dict, molecules):
        d = OrderedDict(sorted(compo_dict.items(), key = lambda x:x[0]))
        with open(outpath + "/Clusters_composition.txt","w") as tmpfile:
            tmpfile.write(f"Cluster Size \t {molecules} : frequency\n\n")
            for k, v in d.items():
                unique_comp = [list(x) for x in set(tuple(x) for x in v)]
                freq = [v.count(comp)/len(v) for comp in unique_comp]
                tmpfile.write(f"  {k}\t\t")
                for cmp, occur in zip(unique_comp, freq):
                    cmp = [str(s) for s in cmp]
                    tmpfile.write(",".join(cmp))
                    tmpfile.write(" : {:.2f}%\t".format(occur*100))
                tmpfile.write("\n\n")
    
    @staticmethod
    def writeDistribution(outpath, cluster_stat):
        TM_eff, N = sum(cluster_stat), len(cluster_stat)
        unique_clusters = sorted(set(cluster_stat))
        freq = [cluster_stat.count(clus)/N for clus in unique_clusters]
        foTM = [cluster_stat.count(clus)*(clus/TM_eff) for clus in unique_clusters]

        with open(outpath+"/SteadyState_distribution.csv", "w", newline='') as tmpfile:
            wf2 = writer(tmpfile)
            wf2.writerow(['Cluster size','frequency','foTM'])
            wf2.writerows(zip(unique_clusters, freq, foTM))
            tmpfile.close()
    
    @displayExecutionTime
    def getSteadyStateDistribution(self, SS_timePoints):
        print("Getting steadystate cluster distribution ...")
        simObj = self.simfileObj
        inpath = simObj.getInpath()
        if self.withMonomer:
            outpath = simObj.getOutpath("Cluster_stat")
        else:
            outpath = simObj.getOutpath("Cluster_stat_wom")
        
        molecules, counts = simObj.getMolecules()
        nf = abs(Decimal(str(self.dt)).as_tuple().exponent) # Number of decimal points to format the clusterTime.csv file
        composition = defaultdict(list) # composition does not track the monomer identity
        #print("nf : ", nf)
        cluster_stat = []
        numRuns = len(self.runs)
        clusTraj = {}
        for j, run in enumerate(self.runs):
            clusDict = {}
            for tp in SS_timePoints:
                try:
                    df = pd.read_csv((inpath + "/data/Run{}/Clusters_Time_{:.{}f}.csv".format(run, tp, nf)), header=None)
                    oligos = self.getSpeciesCount(df, "Size")
                    
                    mol_in_cluster = [self.getSpeciesCount(df, mol) for mol in molecules]
                    for i, clus in enumerate(oligos):
                        comp = [mol_in_cluster[m][i] for m in range(len(mol_in_cluster))]
                        composition[clus].append(comp)
                    
                    if not self.withMonomer:
                        cluster_stat.extend(oligos)
                    else:
                        monomerCount = df[1][0] - len(oligos)
                        clusDict['1'] = str(monomerCount)
                        tmp = {str(c):str(oligos.count(c)) for c in oligos}
                        clusDict.update(tmp)
                        clusTraj[f'Run_{run}'] = clusDict
                        
                        monomers = [1]*monomerCount
                        cluster_stat.extend(monomers)
                        cluster_stat.extend(oligos)
                except:
                    print(f'Run = {run}, timepoint = {tp}, missing')
                    pass
            
            ProgressBar("Progress", (j+1)/numRuns)
        #print(clusTraj)
        self.writeComposition(outpath, composition, molecules)
        self.writeDistribution(outpath, cluster_stat)
        
        with open(outpath + "/Sampling_stat.txt", "w") as tmpfile:
            ss_tp1000 = [(t*1e3) for t in SS_timePoints]
            tmpfile.write("Number of runs : {}\n\n".format(len(self.runs)))
            tmpfile.write(f"Steady state time points (ms): {ss_tp1000}")
       
        with open(outpath + f"/TrajClusterDist.json", "w") as tmf:
            json.dump(clusTraj, tmf,  indent="")
            
        
class InterSiteDistance:
    
    def __init__(self, simfile, site1_name=None, site2_name=None):
        self.simfileObj = ReadInputFile(simfile)
        if (site1_name is None) or (site2_name is None):
            print("Please provide both the site names")
        self.siteName1 = site1_name
        self.siteName2 = site2_name
        tf, dt, dt_data, dt_image = self.simfileObj.getTimeStats()
        timeSeries = np.arange(0, tf+dt_image, dt_image)
        self.timeSeries = timeSeries
        #print(self.timeSeries)
        self.N_frame = round(tf/dt_image)  # number of frames in vierwervile
        
        
    def __repr__(self):
        simfile = self.simfileObj.txtfile.split('/')[-1]
        info = f"Class : {self.__class__.__name__}\nSystem : {simfile}\nSite1 : {self.siteName1}\t\tSite2 : {self.siteName2}"
        return info
    
    @staticmethod
    def findFile(inpath, numRuns, fileName):
        siteIDfile = None
        for run in range(numRuns):
            fname = inpath + f"/Run{run}/{fileName}.csv"
            if os.path.isfile(fname):
                siteIDfile = fname
                break
            else:
                pass
        if siteIDfile is None:
            print(f"Could not locate SiteIDs.csv across {numRuns} runs")
        else:
            return siteIDfile
        
    
    def getSiteIDs(self):
        id1, id2 = None, None 
        inpath = self.simfileObj.getInpath() + "/data"
        numRuns = self.simfileObj.getNumRuns()
        idFile = self.findFile(inpath, numRuns, "SiteIDs")
        #print('idfile: ', idFile)
        with open(idFile, 'r') as infile:
            '''
            r'\b{word}\b' would find the exact 'word' in lines,
            avoiding 'words' or 'word2' or 'word_nv' etc
            '''
            for line in infile:
                #if re.search(self.siteName1, line, re.IGNORECASE):
                if re.search(r'\b{}\b'.format(self.siteName1), line):
                    #print(f'found {self.siteName1} in {line}')
                    id1 = line.split(',')[0]
                    break
            for line in infile:
                #if re.search(self.siteName2, line, re.IGNORECASE):
                if re.search(r'\b{}\b'.format(self.siteName2), line):
                    #print(f'found {self.siteName2} in {line}')
                    id2 = line.split(',')[0]
                    break
        
        print(f'id1: {id1} \t id2: {id2}')
        
        if id1 is None:
            print(f"Invalid site name! Could not find {self.siteName1}")
        elif id2 is None:
            print(f"Invalid site name! Could not find {self.siteName2}")
        else:
            return id1, id2
        
    
    @staticmethod
    def getSiteLocation(viewerfile, siteID):
        pos = []
        vfile = open(viewerfile, 'r')
        for line in vfile.readlines():
            if (re.search("ID",line) and re.search(siteID, line)):
                coors = line.split()
                x,y,z = float(coors[-1]), float(coors[-2]), float(coors[-3])
                pos.append((x,y,z))
        vfile.close()
        return pos
    
    @staticmethod
    def getDistance(pos1, pos2):
        dist = []
        for p1, p2 in zip(pos1, pos2):
            dx,dy,dz = p1[0]-p2[0], p1[1]-p2[1], p1[2]-p2[2]
            d = np.sqrt(dx*dx + dy*dy +dz*dz)
            dist.append(d)
        return dist
    @displayExecutionTime
    def calculate_distance(self, ss_timepoints=[], writeIt=False):
        print(f"Calculating the distance between {self.siteName1} and {self.siteName2} ...")
        dist_stat = []
        id1, id2 = self.getSiteIDs()
        tf, dt, dt_data, dt_image = self.simfileObj.getTimeStats()
        """
        reverse the dt_image string and then find the decimal; 
        0.02->20.0, count = 2
        0.005 -> 500.0, count = 3
        """
        #decimal_count = "{}".format(dt_image)[::-1].find('.') # for formatting timepoints while writing data
        
        inpath = self.simfileObj.getInpath()
        outpath = self.simfileObj.getOutpath("ISD_stat")
        completeTrajCount = 0 # checks if any timepoint gets skipped within a viewerfile
        vfiles =  glob(inpath + "/viewer_files/*.txt")
        IVF = [] # Incomplete Viewer File
        N_traj = len(vfiles)
        
        for index, vfile in enumerate(vfiles):
            pos1 = self.getSiteLocation(vfile, id1)
            pos2 = self.getSiteLocation(vfile, id2)
            dist = self.getDistance(pos1, pos2)
            if len(dist) == self.N_frame + 1:
                dist_stat.append(dist)
                completeTrajCount += 1
            else:
                traj_name = vfile.split('/')[-1]
                IVF.append(traj_name)
            ProgressBar("Progress", (index+1)/N_traj)
        
        if len(IVF) > 0:
            print("incomplete trajectories: ", len(IVF))
            
        dist_arr = np.array(dist_stat)
        mean_dist = np.mean(dist_arr, axis=0, dtype=np.float64)
        ss_index = getSteadyStateIndices(self.timeSeries, ss_timepoints=ss_timepoints)
        SS_dist = dist_arr[:,ss_index]
        
        if writeIt:
            t1000 = [t*1e3 for t in self.timeSeries]
            ss_tp1000 = [t*1e3 for t  in ss_timepoints]
            with open(outpath + f"/{self.siteName1}_{self.siteName2}_ISD_dynamics.csv","w", newline='') as outfile1, open(outpath + f"/{self.siteName1}_{self.siteName2}_ISD_distribution.txt","w") as outfile2:
                wf1 = writer(outfile1,delimiter=',')  # csv writer
                wf1.writerow(['Time(ms)','ISD(nm)'])
                wf1.writerows(zip(t1000, mean_dist))
                
                np.savetxt(outfile2, SS_dist, fmt="%.4e", delimiter=',')
            with open(outpath + "/Sampling_stat.txt","w") as of:
                print("Complete trajectories : ", completeTrajCount)
                of.write(f"Steady state time points (ms) : {ss_tp1000}\n\n")
                of.write(f"Number of complete trajectories : {completeTrajCount}\n\n")
                if len(IVF) > 0:
                    #print("Incomplete trajectories:")
                    of.write("Incomplete trajectories:\n\n")
                    for traj_name in IVF:
                        name = traj_name.split('/')[-1]
                        #print (name)
                        of.write(f"{name}\n\n")
   
            print("\nWriting data is complete\n")

class SingleMoleculeTracking:
    
    def __init__(self, simfile, moleculeName, RadGy_calc=True, MSD_calc=True, coordinate=None):
        self.simfileObj = ReadInputFile(simfile)
        tf, dt, dt_data, dt_image = self.simfileObj.getTimeStats()
        timeSeries = np.arange(0, tf+dt_image, dt_image)
        self.timeSeries = timeSeries
        #self.timeSeries = [int(t*1e3) for t in timeSeries]
        #print(self.timeSeries)
        self.N_frame = round(tf/dt_image)  # number of frames in vierwervile
        self.molName = moleculeName
        self.radGy_calc = RadGy_calc
        self.msd_calc = MSD_calc
        self.coor = coordinate
        print(f"tf = {tf}, dt_image = {dt_image}, N_frame = {self.N_frame}")
        
    def __repr__(self):
        simfile = self.simfileObj.txtfile.split('/')[-1]
        info = f"Class : {self.__class__.__name__}\nSystem : {simfile}\nMolecule: {self.molName}\nRadGy: {self.radGy_calc}\t\tMSD_calc: {self.msd_calc}\t dimension: {self.coor}"
        return info
    
    @staticmethod
    def getMolecularSiteIDs(inpath, numRuns, molName):
        idFile = InterSiteDistance.findFile(inpath, numRuns, "SiteIDs")
        ids = []
        with open(idFile, 'r') as tmpfile:
            for line in tmpfile.readlines():
                if line.split(',')[-1].split()[0] == molName:
                    sid = line.split(',')[0]
                    ids.append(sid)
        return ids
        
    @staticmethod
    def getCoordinates(viewerfile, siteIdList):
        frame_indices = []
        Rg = []
        mcList = []  # coordinates of molecular centroids
        with open(viewerfile, 'r') as tmpfile:
            lines = tmpfile.readlines()
            for i, line in enumerate(lines):
                if re.search("SCENE", line):
                    frame_indices.append(i)
            frame_indices.append(len(lines))
            j = 0
            while(j < len(frame_indices) - 1):
                cur_frame = lines[frame_indices[j]:frame_indices[j+1]]
                j += 1
                coors = []
                for line in cur_frame:
                    if re.search("ID", line):
                        line = line.split()
                        if line[1] in siteIdList:
                            coors.append((float(line[-1]), float(line[-2]), float(line[-3])))
                        
                N = len(coors)
                xs,ys,zs = 0,0,0
                for x,y,z in coors:
                    xs += x
                    ys += y
                    zs += z
                xm,ym,zm = xs/N, ys/N, zs/N  # coordinate of molecular centroid
                mcList.append((xm,ym,zm))
                
                gen_coor = np.array([np.sqrt(x*x+y*y+z*z) for (x,y,z) in coors])
                cop = np.mean(gen_coor)  # center of position
                Rp2 = sum((gen_coor-cop)**2)/len(gen_coor) # RadGy_pos_squared, array operation
                Rg.append(np.sqrt(Rp2))
        return Rg, mcList
    
    @displayExecutionTime
    def getMolecularTrajectory(self, ss_timepoints=[], writeIt=True):
        print("Tracking the molecular trajectory...")
        simObj = self.simfileObj
        inpath = simObj.getInpath()
        outpath = simObj.getOutpath("smTracking_stat")
        
        siteInpath = inpath + "/data"
        numRuns = simObj.getNumRuns()
        siteIdList = self.getMolecularSiteIDs(siteInpath, numRuns, self.molName)
        
        Rg_stat, mcCoor_stat = [], []
        vfiles =  glob(inpath + "/viewer_files/*.txt")
        #vfiles =  glob(inpath + "/test/*.txt")
        IVF = [] # Incomplete Viewer File
        N_traj = len(vfiles)
        count_traj = 0
        for i, vfile in enumerate(vfiles):
            Rg_traj, mcList = self.getCoordinates(vfile, siteIdList)
            #print(f"Rg_traj length : {len(Rg_traj)} \t N_frame = {self.N_frame}")
            if len(Rg_traj) == self.N_frame + 1:
                count_traj += 1
                Rg_stat.append(Rg_traj)
                mcCoor_stat.append(mcList)
            else:
                IVF.append(vfile)
            ProgressBar("Progress", (i+1)/N_traj)
        
        t1000 = [t*1e3 for t in self.timeSeries]
        ss_tp1000 = [t*1e3 for t in ss_timepoints]
        
        with open(outpath + "/Sampling_stat.txt","w") as of:
            print("Complete trajectories : ", count_traj)
            of.write(f"Steady state time points (ms) : {ss_tp1000}\n\n")
            of.write(f"Number of complete trajectories : {count_traj}\n\n")
            if len(IVF) > 0:
                print("Incomplete trajectories:")
                of.write("Incomplete trajectories:\n\n")
                for traj_name in IVF:
                    name = traj_name.split('/')[-1]
                    print (name)
                    of.write(f"{name}\n\n")
        
        if self.msd_calc:
            print("Calculating the molecular diffusion....")
            SD_traj = []   # squared displacement
            dist_func_3d = lambda p1, p2: (p2[0]-p1[0])**2 + (p2[1]-p1[1])**2 + (p2[2]-p1[2])**2
            dist_func_1d = lambda p1, p2: (p2[0]-p1[0])**2 
            
            """
            r = (x, y, z)
            mcCoor_stat = [[r_0t0, r_0t1, r_0t2, ..., r_0tn], ..., [r_nt0, r_nt1, ..., r_ntn]]
            disp = dr**2 = (xt-x0)**2 + (yt-y0)**2 + (zt-z0)**2 
            msd = mean of disp over n independent trajectories
            """
            for traj in mcCoor_stat:
                dr2_traj = []
                tn = 0
                while tn < len(traj):
                    if self.coor.upper() == '3D':
                        dr2 = dist_func_3d(traj[tn], traj[0])
                    elif self.coor.upper() == '1D':
                        dr2 = dist_func_1d(traj[tn], traj[0])
                        
                    dr2_traj.append(dr2)
                    tn += 1
                SD_traj.append(dr2_traj)
            
            MSD_traj = np.mean(np.array(SD_traj), axis=0)
            
                
            if writeIt:
                with open(outpath + f"/{self.molName}_MSD.csv","w", newline='') as tmpf:
                    wf = writer(tmpf, delimiter=',')  # csv writer
                    wf.writerow(['Time(s)','MSD(nm^2)'])
                    # timepoints have to be in 'seconds' unit in order to get correct D unit
                    wf.writerows(zip(self.timeSeries, MSD_traj))
                print("... done writing!")
             
        
        if self.radGy_calc:
            print("Calculating Radius of Gyration....")
            Rg_arr = np.array(Rg_stat)
            mean_Rg = np.mean(Rg_arr, axis=0, dtype=np.float64)
            ss_index= getSteadyStateIndices(self.timeSeries, ss_timepoints=ss_timepoints)
            SS_Rg = Rg_arr[:,ss_index]
        
            if writeIt:
                with open(outpath + f"/{self.molName}_RadGy_dynamics.csv","w", newline='') as outfile1, open(outpath + f"/{self.molName}_RadGy_distribution.txt","w") as outfile2:
                    wf1 = writer(outfile1,delimiter=',')  # csv writer
                    wf1.writerow(['Time(ms)','RadGy(nm)'])
                    wf1.writerows(zip(t1000, mean_Rg))
                    np.savetxt(outfile2, SS_Rg, fmt="%.4e", delimiter=',')
                print("... done writing!")
            

class RunTimeStatistics:
    
    def __init__(self, txtfile, runs=None):
        self.simObj = ReadInputFile(txtfile)
        if runs == None:
            numRuns = self.simObj.getNumRuns()
            self.Runs = [int(i) for i in range(numRuns)]
        else:
            self.Runs = runs
        
    def __repr__(self):
        simfile = self.simObj.txtfile.split('/')[-1]
        info = f"Class : {self.__class__.__name__}\nSystem : {simfile}"
        return info
    
    @staticmethod
    def getRunTime(runfile):
        runTime, timeUnit = None, None
        with open(runfile, 'r') as tmpfile:
            line = tmpfile.readlines()[0]
            tps = re.findall(r"[-+]?\d*\.\d+|\d+", line)
            if len(tps) == 1:
                runTime = float(tps[0])
                timeUnit = "sec"
            elif len(tps) == 2:
                runTime = (float(tps[0])*60 + float(tps[-1]))/60
                timeUnit = "min"
            elif len(tps) == 3:
                runTime = (float(tps[0])*3600 + float(tps[1])*60 + float(tps[2]))/3600
                timeUnit = "hour"
            elif len(tps) == 4:
                runTime = (float(tps[0])*24*3600 + float(tps[1])*3600 + float(tps[2])*60 + float(tps[3]))/3600
                timeUnit = "hour"
        return runTime, timeUnit
     
    @displayExecutionTime 
    def getStat(self, fontsize=16):
        print("Getting runtime stats ...")
        seqNo, RT, units = [], [], []
        inpath = self.simObj.getInpath() + "/data"
        outpath = self.simObj.getOutpath("RunTime_stat")
        N_runs = len(self.Runs)
        for i, run in enumerate(self.Runs):
            try:
                rf = inpath + f"/Run{run}/RunningTime.txt"
                rt, unit = self.getRunTime(rf)
                seqNo.append(run)
                RT.append(rt)
                units.append(unit)
                ProgressBar("Progress", (i+1)/N_runs)
            except:
                pass
        #print(units)
        if len(set(units)) == 1: # when all units are of same type (sec/min/hour)
            if units[0] == "sec":
                plt.ylabel("Runtime (secs)", fontsize=fontsize)
            elif units[0] == "min":
                plt.ylabel("Runtime (mins)", fontsize=fontsize)
            elif units[0] == "hour":
                plt.ylabel("Runtime (hours)", fontsize=fontsize)
            plt.xlabel("Run Sequence", fontsize=fontsize)
            plt.scatter(seqNo, RT, color='b')
            with open(outpath + '/runTimeInfo.csv', 'w', newline='') as tmpf:
                wf = writer(tmpf)
                wf.writerows(zip(seqNo, RT))
            simName = self.simObj.txtfile.split('/')[-1].replace(".txt","")
            plt.title(simName)
            plt.savefig(outpath + "/runTime.png", dpi=100)
            plt.close()
            print("Done plotting the stats ...!")
        else:
            print("Time units are different! Can't plot the runtime stats")
        
            




import os
import importlib.resources as pkg_resources
import subprocess
import platform

def simulate(model, runs=1):
    
    os_name = platform.system()

    if os_name == 'Windows':
        executable_name = 'langevin-windows-latest'
        executable_extension = '.exe'
    elif os_name == 'Darwin':
        if platform.version().split('.')[0] == '14':
            executable_name = 'langevin-macos-14'
        else:
            executable_name = 'langevin-macos-13'
        executable_extension = '.app'
    else:
        executable_name = 'langevin-ubuntu-latest'
        executable_extension = ''

    with pkg_resources.path('saladpy', executable_name) as executable_path:
        executable_path = f"{executable_path}{executable_extension}"
        model_path = os.path.abspath(model)
        print(f'Model "{model}" is running') 
        raw_s = r'{}'.format(model_path)
        for i in range(runs):
            subprocess.run([executable_path, 'simulate', raw_s, str(i)])
        output_model_path = os.path.join(model_path[:-4] + '_FOLDER', os.path.split(model_path)[1])
        with open(output_model_path, mode='a') as file:
            file.write(f'Runs: {str(runs)}')
        print(f'Simulation complete, results can be found here: {model_path[:-4]}_FOLDER')

#simulate(r'C:\Users\cpero\Downloads\test_output\Simulation0_SIM.txt')



from .ssalad_ClusterAnalysis import *
from .Molclustpy_visualization_funcitons import *
from .Composition_calculator import composition_calc
from .Format import format

def plot(search_directory, times, special_clusters=[], width=0.1, alpha=0.6, legend_right=True):
    
    input_file, rounded_times, title_str = format(search_directory, times, 'cluster')

    ca = ClusterAnalysis(input_file)
    ca.getMeanTrajectory(SingleTraj=False)
    ca.getSteadyStateDistribution(SS_timePoints=rounded_times)

    composition_calc(search_directory, title_str, special_clusters, width=width, alpha=alpha, legend_right=legend_right)


import numpy as np
from .Reader_functions import *
from .data_locator import *
from .input_file_extraction import *
from .DrawingTool import Draw_2D_Molecule, ReadSimFile
from SpringSaLaDpy.data_locator import find_txt_file


def Describe_input_file(search_directory, search_term='', links=False, reactions=False, kinetics=False, drawings=False):
    molecules, split_file = read_input_file(search_directory, search_term)

    txtfile = find_txt_file(search_directory)
    saveImage = False
    simfile = ReadSimFile(txtfile)
    #outpath = simfile.getOutPath()
    outpath = ''
    molNames, molcounts, SiteList, LinkList = simfile.getSiteStats()

    #print(SiteList)
    LineWidth = 0.75

    print('Molecules:\n')

    #Molecule Processing
    multi_state_types = []
    multi_state_sites = []
    for mol_num, item in enumerate(molecules):   
        #General Info
        typStr = ''
        split_general_info = item[0].split()
        site_types = int(split_general_info[6])
        for j in range(2,2+site_types):
            type_split = item[j].split()
            type_name = type_split[2]
            states = [i for i in type_split if '"' in i]
            site_name = states.pop(0)
            typStrAddition = ''
            if len(states) == 1:
                pass
            else:
                for state in states:
                    typStrAddition = typStrAddition + f'{state[1:-1]}, '
                typStrAddition = typStrAddition[:-2]
                typStrAddition = f' ({len(states)} possible states: {typStrAddition})'
                multi_state_types.append(type_name[1:-1])
            typStr = f'{typStr}{type_name[1:-1]}{typStrAddition}, '
            if kinetics:
                initial_count = f' (Initial Count: {split_general_info[4]})'
            else:
                initial_count = ''
        print(f'MOLECULE: {split_general_info[1]}{initial_count}')
        
        if drawings:
            mol, site, link = tuple(zip(molNames, SiteList, LinkList))[mol_num]
            mol2D = Draw_2D_Molecule(mol,site,link,outpath)
            mol2D.displayMolecule(LineWidth, saveImage)

        print(f'This molecule has {split_general_info[6]} site types: {typStr[:-2]}')

        #Sites
        total_sites = int(split_general_info[8])
        site_data = []
        for i in range(3 + site_types, 3 + site_types + 3*total_sites, 3):
            site_split1 = item[i].split()
            site_split2 = item[i+1].split()
            if len(site_split2) > 11:
                multi_state_sites.append(int((i-site_types)/3) - 1)
            site = [f'{site_split1[0]} {site_split1[1]}', site_split1[3], site_split2[2][1:-1], site_split2[8]]
            site_data.append(site)

        #Sites (default ordering)
        '''
        print(f'\nIt consists of {total_sites} connected sites:')
        for site in site_data:
            print(f'Site {site[0][5:]} ({site[1]}) of type "{site[2]}"')
            pass
        '''

        #Sites (grouped by type)
        sites_reordered = []
        site_type_names = []  
        for site in site_data:
            if site[2] not in site_type_names:
                site_type_names.append(site[2])
                #Site type name, Numpy Array(#Intracellular, #Extracellular, #Membrane), [list of sites], color 
                sites_reordered.append([site[2], location2vec(site[1]), [site[0]], site[3].lower().capitalize()])
            else:
                for reordered_site in sites_reordered:
                    if site[2] == reordered_site[0]:
                        reordered_site[1] = reordered_site[1] + location2vec(site[1])
                        reordered_site[2].append(site[0])
        
        print(f'\nIt consists of {total_sites} connected sites:')
        for reordered_site in sites_reordered:
            location_list = list(reordered_site[1])
            site_list = list(reordered_site[2])
            for index in location_list:
                addStr = ''
                if location_list[0] != 0:
                    addStr = addStr + f'{location_list[0]} {reordered_site[3]} intracellular, '
                if location_list[1] != 0:
                    addStr = addStr + f'{location_list[1]} {reordered_site[3]} extracellular, '
                if location_list[2] != 0:
                    addStr = addStr + f'{location_list[2]} {reordered_site[3]} membrane, '
            addStr2 = ''
            for index in site_list:
                addStr2 = addStr2 + 'Site' + index[4:] + ', '
            if ',' in addStr2[:-2]:
                print(f'Type {reordered_site[0]}: {addStr[:-2]} sites ({addStr2[:-2]})')
            else:
                print(f'Type {reordered_site[0]}: {addStr[:-2]} site ({addStr2[:-2]})')        
        
        total_links = int(split_general_info[10])
        link_data = []
        for i in range(4 + site_types + 3*total_sites, 4 + site_types + 3*total_sites + total_links):
            split_link = item[i].split()
            link = [split_link[2], split_link[5]]
            link_data.append(link)
        
        #[[link index, list of links],[link index, list of links],...]
        link_lists = []
        for site in site_data:
            split_site = site[0].split()
            link_count_init = [split_site[1], []]
            link_lists.append(link_count_init)
        for link_list in link_lists:
            for link in link_data:
                if link_list[0] == link[0]:
                    link_list[1].append(link[1])
                elif link_list[0] == link[1]:
                    link_list[1].append(link[0])
                else:
                    pass

        link_lists_remaining = link_lists

        #Links (no repeated entries)
        '''
        print(f'\nThis molecule had {total_links} total links between sites:')
        while True:
            record_length = 0
            for link_list in link_lists:
                if len(link_list[1]) > record_length:
                    record_length = len(link_list[1])
                    record_loc = int(link_list[0])
            if record_length == 0:
                break

            site_str = ''
            for site in link_lists_remaining[record_loc][1]:
                site_str = f'{site_str} site {site},'
            
            print(f'Site {record_loc} is connected to{site_str[:-1]}')
            for link_list_remaining in link_lists_remaining:
                link_list_remaining[1] = [i for i in link_list_remaining[1] if i != str(record_loc)] 
            link_lists_remaining[record_loc][1] = []
        print('\n')
        '''


        #All links for all sites
        if links:
            print(f'\nIt has {total_links} total links between sites:')
            for link in link_lists:
                site_str = ''
                for site in link[1]:
                    site_str = f'{site_str} site {site},'
                print(f'Site {link[0]} is connected to{site_str[:-1]}')
        print('------------------------------------------------------------------------------------------------------------\n')

    if reactions:
        transition_reactions, allosteric_reactions, binding_reactions = read_reactions(split_file)
        print('************************************************************************************************************\n')
        print('Reaction Rules')
        #State Transition Reactions
        if len(transition_reactions) != 0:
            print(f'\nState Transition Reactions: {len(transition_reactions)}')
            for transition_reaction in transition_reactions:
                split_reaction = transition_reaction.split('\'')
                split_condition = split_reaction[10].split()
                condition = split_condition[-1]

                if split_reaction[13] not in multi_state_types:
                    state = ''
                elif split_reaction[15] == 'Any_State':
                    state = ' in any state'
                else:
                    state = ' of state ' + split_reaction[15]
                
                if condition == 'None':
                    condition_str = ''
                elif condition == 'Free':
                    condition_str = f'provided that it is free'
                else:
                    condition_str = f'provided that it is bound to site {split_reaction[13]} of a molecule {split_reaction[11]}{state}'

                if kinetics:
                    rate = f'at a rate of {split_reaction[10].split()[1]} s-1 '
                else:
                    rate = ''
                print(f'A site of type {split_reaction[5]} in a {split_reaction[3]} molecule can change its state from {split_reaction[7]} to {split_reaction[9]} {rate}{condition_str}')

        #Allosteric Reactions
        if len(allosteric_reactions) != 0:
            print(f'\nAllosteric Reactions: {len(allosteric_reactions)}')
            for allosteric_reaction in allosteric_reactions:
                split_reaction = allosteric_reaction.split('\'')
                split_allosteric_site = split_reaction[-3].split()

                if int(split_allosteric_site[3]) not in multi_state_sites:
                    condition = ''
                else:
                    condition = f'provided that Site {split_allosteric_site[3]} of the same molecule is in state {split_reaction[-2]}'

                if kinetics:
                    rate = f'at a rate of {split_reaction[8].split()[1]} s-1 '
                else:
                    rate = ''
                print(f'{split_reaction[4][3:-3]} of molecule {split_reaction[3]} can change its state from {split_reaction[5]} to {split_reaction[7]} {rate}{condition}')

        #Biomolecular Binding Reactions
        if len(binding_reactions) != 0:
            print(f'\nBiomolecular Binding Reactions: {len(binding_reactions)}')
            for binding_reaction in binding_reactions:
                split_reaction = binding_reaction.split('\'')
                
                if split_reaction[5] not in multi_state_types:
                    state1 = ''
                elif split_reaction[7] == 'Any_State':
                    state1 = ' (in any state)'
                else:
                    state1 = f' (of state {split_reaction[7]})'
                
                if split_reaction[11] not in multi_state_types:
                    state2 = ''
                elif split_reaction[13] == 'Any_State':
                    state2 = ' (in any state)'
                else:
                    state2 = f' (of state {split_reaction[13]})'

                if kinetics:
                    on_rate = f'\n\tOn rate: {split_reaction[14].split()[1]} uM-1.s-1,'
                    off_rate = f'off rate: {str(split_reaction[14].split()[3])} s-1'
                    #if float(split_reaction[14].split()[3]) != 0.0:
                else:
                    on_rate = off_rate = ''
                print(f'Site {split_reaction[5]} of molecule {split_reaction[3]}{state1} can bind to site {split_reaction[11]} of {split_reaction[9]}{state2} {on_rate} {off_rate}')




def times_2_title(times):
    if len(times) > 1:
            title_str = ' at: '
            for time in times:
                title_str = f'{title_str}{str(time)}, '
            title_str = title_str[:-2]
    else:
        title_str = ' at '
        title_str = f'{title_str}{str(times[0])}'
    title_str = title_str + ' s'
    return title_str


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 16 18:43:43 2021

@author: Ani Chattaraj
"""

from .DataPy import ReadInputFile, InterSiteDistance, ProgressBar, displayExecutionTime
import re, json, pickle
import numpy as np
import networkx as nx
from glob import glob
import matplotlib.pyplot as plt
from csv import writer
from numpy import pi, array
from collections import defaultdict, OrderedDict, namedtuple, Counter
import os
from SpringSaLaDpy.data_locator import data_file_finder

font = {'family' : 'Arial',
        'size'   : 16}

plt.rc('font', **font)


def connected_component_subgraphs(G):
    for c in nx.connected_components(G):
        yield G.subgraph(c)
# pos = center of mass, radius = Radius of gyration, 
#density = sites/volume
Cluster = namedtuple('Cluster', ['pos', 'radius', 'density'])

class ClusterDensity:

    def __init__(self, txtfile, ss_timeSeries):
        self.simObj = ReadInputFile(txtfile)
        tf, dt, dt_data, dt_image = self.simObj.getTimeStats()
        self.box = self.simObj.getBoundingBox()
        self.ss_timeSeries = ss_timeSeries
        inpath = self.simObj.getInpath() + "/data"
        numRuns = self.simObj.getNumRuns()
        self.N_frames = int(tf/dt_image)
        self.inpath = inpath
        if numRuns == 0:
            self.numRuns = 25
        else:    
            self.numRuns = numRuns

    def __repr__(self):
        simfile = self.simObj.txtfile.split('/')[-1]
        info = f"Class : {self.__class__.__name__}\nSystem : {simfile}"
        return info

    @staticmethod
    def getMolIds(molfile, molName):
        molDict = {}
        mIds = []
        with open(molfile, 'r') as tmpfile:
            for line in tmpfile:
                line = line.strip().split(',')
                if line[-1] == molName:
                    mIds.append(line[0])
        molDict[molName] = mIds
        return molDict

    @staticmethod
    def getSiteIds(sitefile, molName):
        siteDict = {}
        sIds = []
        with open(sitefile, 'r') as tmpfile:
            for line in tmpfile:
                line = line.strip().split(',')
                if line[-1].split()[0] == molName:
                    sIds.append(line[0])
        siteDict[molName] = sIds
        return siteDict

    @staticmethod
    def getRevDict(myDict):
        # k,v = key : [val1, val2, ...]
        revDict = {}
        for k,v in myDict.items():
            for val in v:
                revDict[val] = k
        return revDict
    @staticmethod
    def splitArr(arr, n):
        subArrs = []
        if (len(arr)%n == 0):
            f = int(len(arr)/n)
            i = 0
            while (i < len(arr)):
                sub_arr = arr[i : i+f]
                i += f
                subArrs.append(sub_arr)
            return subArrs
        else:
            print(f"Can't split the given array (length = {len(arr)}) into {n} parts")

    def mapSiteToMolecule(self):
        sIDfile = InterSiteDistance.findFile(self.inpath, self.numRuns, "SiteIDs")
        mIDfile = InterSiteDistance.findFile(self.inpath, self.numRuns, "MoleculeIDs")
        molName, molCount = self.simObj.getMolecules()
        molIDs, siteIDs = {}, {}
        for mol in molName:
            molIDs.update(self.getMolIds(mIDfile, mol))
            siteIDs.update(self.getSiteIds(sIDfile, mol))
        spmIDs = {} # sites per molecule
        for mol, count in zip(molName, molCount):
            #molDict = {}
            arr = self.splitArr(siteIDs[mol], count)
            mol_ids = molIDs[mol]
            d = dict(zip(mol_ids, arr))
            spmIDs.update(d)
        rev_spm = self.getRevDict(spmIDs)
        return rev_spm

    @staticmethod
    def getBindingStatus(frame):
        linkList = []
        posDict = {}
        colorDict = {}
        radDict = {}
        for curline in frame:
            if re.search("ID", curline):
                line = curline.split()
                posDict[line[1]] = [float(line[4]),float(line[5]), float(line[6])] # order = x,y,z
                colorDict[line[1]] = line[3]
                radDict[line[1]] = float(line[2])
                #IdList.append(line.split()[1])

            if re.search("Link", curline):
                line = curline.split()
                linkList.append((line[1], line[3]))
        return posDict, linkList, colorDict, radDict

    @staticmethod
    def createGraph(IdList, LinkList):
        G = nx.Graph()
        G.add_nodes_from(IdList)
        G.add_edges_from(LinkList)
        return G
    
    @staticmethod
    def createMultiGraph(LinkList):
        MG = nx.MultiGraph()
        MG.add_edges_from(LinkList)
        return MG

    @staticmethod
    def getFrameIndices(viewerfile):
        frame_indices = []
        tps = []
        with open(viewerfile, 'r') as tmpfile:
            lines = tmpfile.readlines()
            for i, line in enumerate(lines):
                if re.search("SCENE", line):
                    frame_indices.append(i)
                    tp = lines[i+1].split()[-1]
                    tps.append(tp)

            frame_indices.append(len(lines))

        return tps, frame_indices
    
    def getSteadyStateFrameIndices(self, viewerfile):
        frame_indices = []
        ss_indices = [] 
        tps = []
        index_pairs = []
        with open(viewerfile, 'r') as tmpfile:
            lines = tmpfile.readlines()
            for i, line in enumerate(lines):
                if re.search("SCENE", line):
                    frame_indices.append(i)
                    tp = lines[i+1].split()[-1]
                    tps.append(tp)
                    if any([np.isclose(float(tp), t) for t in self.ss_timeSeries]):
                        ss_indices.append(i)
            frame_indices.append(len(lines))
        for ii, elem in enumerate(frame_indices):
            if elem in ss_indices:
                index_pairs.append((elem, frame_indices[ii+1]))
        return tps, index_pairs

    @staticmethod
    def calc_RadGy(posList, massList, cubic_com=True):
        if cubic_com:
            massList = massList.reshape(-1, 1)
            weighted_array = posList * massList
            weighted_sum = np.sum(weighted_array, axis=0)
            sum_weights = np.sum(massList)
            center = weighted_sum / sum_weights
        else:
            # posList = N,3 array for N sites
            center = np.mean(posList, axis=0) # center of cluster
        
        Rg2 = np.mean(np.sum((posList - center)**2, axis=1))
        rmax2 = np.max(np.sum((posList - center)**2, axis=1), axis=0)
        return center, np.sqrt(Rg2), np.sqrt(rmax2)
    
    @staticmethod
    def calc_zagreb_indices(MG, node_list):
        # MG: MULTI-GRAPH Object # multiple edges allowed between two nodes 
        d1List = []
        d2List = []
        nodes = list(MG.nodes())
        specific_mol = []
        #links = [(nodes[i], nodes[j]) for i in range(len(nodes)) for j in range(len(nodes)) if i<j]
        for n in nodes:
            d1List.append(MG.degree(n))
        for n1,n2 in set(MG.edges()):
            d2List.append((MG.degree(n1), MG.degree(n2)))
        d1Arr = array(d1List)
        
        specific_nodes = [node for node in nodes if node in node_list]

        for node in specific_nodes:
            specific_mol.append(MG.degree(node))

        # M1, M2: first and second Zagreb indices
        M1 = sum(d1Arr**2)
        M2 = sum([d1*d2 for d1,d2 in d2List])
        
        if node_list == []:
            specific_mol = d1Arr

        return M1, M2, d1Arr, specific_mol
        

    def getClusterDensity(self, viewerfile, cs_thresh=1, molecule_list=[], cubic_com=True):
        # cluster size,  radius of gyration
        # M1, M2: Zagreb indices
        csList, RgList, rmaxList, M1List, M2List = [], [], [], [], []
        
        mtp_cs, mtp_rg, mtp_rmax = [], [], [] # mtp: multi timepoint stat
        
        # MCL: molecular cross linking (number of bonds per molecule)  
        MCL = []
        msm = self.mapSiteToMolecule()
        tps, index_pairs = self.getSteadyStateFrameIndices(viewerfile)

        run_num = os.path.split(viewerfile)[1].split('_')[-1][3:-4]
        search_directory = os.path.join(os.path.split(os.path.split(viewerfile)[0])[0])
        molecule_name_file = data_file_finder(search_directory=search_directory, path_list=['data', f'Run{run_num}'], search_term='MoleculeIDs.csv')
        name_to_IDs = {}
        with open(molecule_name_file, 'r') as file:
            lines = file.readlines()
            for line in lines:
                name = line.split(',')[1][:-1]
                ID_num = line.split(',')[0]
                if name not in name_to_IDs.keys():
                    name_to_IDs[name] = [ID_num]
                else:
                    name_to_IDs[name].append(ID_num)

        node_list = [] 
        if molecule_list != []:
            for molecule in molecule_list:
                node_list.extend(name_to_IDs[molecule])

        path = os.path.join(os.path.split(molecule_name_file)[0], 'molecule_links.txt')
        open(path, 'w').close()

        with open(viewerfile, 'r') as infile:
            lines = infile.readlines()
            for i,j in index_pairs:
                #i,j = index_pairs[-1]
                current_frame = lines[i:j]
                cs_frame, rg_frame, rmax_frame = [], [], [] # clusters in current frame
                posDict, Links, _, radDict = self.getBindingStatus(current_frame)
                Ids = [_ for _ in posDict.keys()]
                #mIds, mLinks = [msm[k] for k in Ids], [(msm[k1], msm[k2]) for k1,k2 in Links]
                sG = self.createGraph(Ids, Links)
                #mG = self.createGraph(mIds, mLinks)
                
                with open(path, 'a') as outfile:
                    outfile.write(f'\n\n{current_frame[0]}')
                    outfile.write(current_frame[1])
                outfile.close()

                #G.subgraph(c) for c in connected_components(G)
                count = 0
                for sg in connected_component_subgraphs(sG):
                    mLinks = [(msm[k1], msm[k2]) for k1, k2 in sg.edges()]
                    # connection between two different molecules                    
                    bonds = [(m1,m2) for m1,m2 in mLinks if m1 != m2]
                    
                    if len(bonds) > 0:
                        count += 1
                        with open(path, 'a') as outfile:
                            outfile.write(f'\nCluster {count}')
                            for bond in bonds:
                                outfile.write(f'\nBOND\t{bond[0]}\t:\t{bond[1]}')
                        outfile.close()

                    MG = self.createMultiGraph(bonds)

                    # cluster size (number of molecules)
                    cs = len(MG.nodes())

                    if cs > cs_thresh:
                        M1, M2, dArr, specific_mols = self.calc_zagreb_indices(MG, node_list=node_list)
                        MCL.extend(specific_mols)
                       
                        sites = list(sg.nodes)
                        
                        posList = np.array([posDict[s] for s in sites])
                        massList = np.array([float(radDict[s])**3 for s in sites])
                        
                        center, Rg, rmax = self.calc_RadGy(posList, massList, cubic_com=cubic_com)
                        
                        cs_frame.append(cs)
                        rg_frame.append(Rg)
                        rmax_frame.append(rmax)
                        
                        csList.append(cs)
                        RgList.append(Rg)
                        rmaxList.append(rmax)
                        M1List.append(M1)
                        M2List.append(M2)
                
                mtp_cs.append(cs_frame)
                mtp_rg.append(rg_frame)
                mtp_rmax.append(rmax_frame)
                        
        return [csList, RgList, rmaxList, M1List, M2List], MCL, mtp_cs, mtp_rg, mtp_rmax
    
    @staticmethod  
    def plotRg(csList, RgList, time_str):
        fig, ax = plt.subplots(figsize=(5,3))
        ax.scatter(csList, RgList, color='k', s=4)
        ax.set_xlabel('Cluster size (molecules)')
        ax.set_ylabel('Radius of Gyration (nm)')
        ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True, min_n_ticks=1))
        plt.title('Radius of Gyration Scatter Plot' + time_str)
        plt.show() 
    
    @staticmethod
    def plotBondsPerMolecule(countDict, time_str, mol_str):
        fig, ax = plt.subplots(figsize=(5,3))
        bonds, freq = countDict.keys(), countDict.values()
        ax.bar(bonds, freq, width=0.3, color='grey')
        ax.set_xlabel('Bonds per molecule')
        ax.set_ylabel('Frequency')
        ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True, min_n_ticks=1))
        plt.title(f'Bonds per Molecule{mol_str} Histogram{time_str}')
        plt.show()
        
    @displayExecutionTime
    def getCD_stat(self, cs_thresh=1, title_str='', hist=False, indices=[], cubic_com=True):
        # collect statistics at the last timepoint
        sysName = self.inpath.split('/')[-2].replace('_SIM_FOLDER','')
        print('\nSystem: ', sysName)
        print("Calculating Cluster Density ...")
      
        outpath = self.simObj.getOutpath("BF_stat")
        vfiles = glob(self.simObj.getInpath() + "/viewer_files/*.txt")[:]

        N_traj = len(vfiles)

        header = 'Cluster size, Rg (nm), M1, M2'
        MCL_stat = []
        cs_tmp, rg_tmp = [], []

        print(vfiles)

        search_directory = os.path.join(os.path.split(os.path.split(vfiles[0])[0])[0])
        molecule_name_file = data_file_finder(search_directory=search_directory, path_list=['data', f'Run{0}'], search_term='MoleculeIDs.csv')
        names = []
        with open(molecule_name_file, 'r') as file:
            lines = file.readlines()
            for line in lines:
                name = line.split(',')[1][:-1]
                if name not in names:
                    names.append(name)

        molecule_list = []

        for index in indices:
            molecule_list.append(names[index])

        for i, vfile in enumerate(vfiles):
            res, MCL, mtp_cs, mtp_rg, mtp_rmax = self.getClusterDensity(vfile, cs_thresh=cs_thresh, molecule_list=molecule_list, cubic_com=cubic_com)
            #print(array(mtp_rg))
            MCL_stat.extend(MCL)
            cs_tmp.extend(mtp_cs)
            rg_tmp.extend(mtp_rg)
            
            if cs_thresh == 1:
                runNum = vfile.split('_')[-1] # Run0.txt
                #np.savetxt(outpath + '/cs_' + runNum , array(mtp_cs))
                with open(outpath + '/cs_' + runNum, 'w') as of:
                    for item in mtp_cs:
                        of.write(f'{item}')
                        of.write('\n')
                
                with open(outpath + '/rg_' + runNum, 'w') as of:
                    for item in mtp_rg:
                        of.write(f'{item}')
                        of.write('\n')
            
            ProgressBar("Progress", (i+1)/N_traj)

        counts_norm = {k: (MCL_stat.count(k)/len(MCL_stat)) for k in set(MCL_stat)}
        
        if cs_thresh == 1:
            fName = '/Bonds_per_single_molecule.csv'
        else:
            fName = f'/Bonds_per_single_molecule_cs_gt_{cs_thresh}.csv'
            
        with open(outpath + fName, "w", newline='') as of:
            obj = writer(of)
            obj.writerow(['BondCounts','frequency'])
            obj.writerows(zip(counts_norm.keys(), counts_norm.values()))
        
        csList = np.concatenate(cs_tmp).ravel().tolist()
        rgList = np.concatenate(rg_tmp).ravel().tolist()
        if hist:
            print('Molecules:')
            all_indices = []
            for i, name in enumerate(names):
                all_indices.append(i)
                print(f'{i}: {name}')
            print('\nList of Indices:')
            print(all_indices)

            add_str = ''
            for index in indices:
                add_str = add_str + f'{names[index]}, '
            add_str = add_str[:-2]
            if add_str != '':
                mol_str = f' ({add_str})'
            else:
                mol_str = ''

            self.plotBondsPerMolecule(counts_norm, title_str, mol_str)
        else:
            self.plotRg(csList, rgList, title_str)

       
'''
files = glob('C:/Users/chatt/Desktop/pytest/springsalad/test_dataset/A5_B5_flex_3nm_2nm_count_40_SIM_FOLDER/A5_B5_flex_3nm_2nm_count_40_SIM.txt')        

for txtfile in files[:]:
    cd = ClusterDensity(txtfile, ss_timeSeries=[ 0.02, 0.04])
    cd.getCD_stat(cs_thresh=1)
    
'''








# -*- coding: utf-8 -*-
"""
Created on Wed Feb  6 21:37:50 2019

@author: Ani Chattaraj
"""

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import os
from math import atan, sin, cos
from collections import namedtuple
import re
import numpy as np

class Draw_2D_Molecule:
    def __init__(self, molName, SiteList, LinkList, outPath):
        self.molecularName = molName
        self.siteList = SiteList
        self.linkList = LinkList
        self.outPath = outPath
    
    @staticmethod
    def draw_circle(SiteObj):
        """
        modify the color string if defined site-color is invalid
        in CSS-style matplotlib implementation (like 'dark_violet') 
        LIME_GREEN : limegreen; Dark_violet : darkviolet
        """
        coor, r = SiteObj.coordinates, SiteObj.radius
        x, y = coor[2], coor[1]
        try:
            if SiteObj.color.lower() == 'white':
                circle = plt.Circle((x,y), radius=r, facecolor='lightsteelblue')
                print("White site color encountered; replaced with lightsteelblue")
            else:
                circle = plt.Circle((x,y), radius=r, facecolor=SiteObj.color)
        except:
            color = SiteObj.color.replace("_","") # "dark_violet" has to be in "darkviolet" format
            circle = plt.Circle((x,y), radius=r, facecolor=color)  
        
        return plt.gca().add_patch(circle)
    
    
    def draw_line(self, SiteObj1, SiteObj2, width):
        """
        take a pair of sites ; draw a line between the surface (excluding the radius)
        For non-linear molecules, calculates the slope between two spheres and
        place the line points along that direction.
        """
        c1, r1 = SiteObj1.coordinates, SiteObj1.radius
        c2, r2 = SiteObj2.coordinates, SiteObj2.radius
        x1, y1 = c1[2], c1[1]
        x2, y2 = c2[2], c2[1]
        
        try:
            slope = atan((y2-y1)/(x2-x1))  # in radians
            
            # pn: peripherial coordinates of site n
            p1_x, p1_y = x1 + r1 * cos(slope), y1 + r1 * sin(slope)  
            p2_x, p2_y = x2 - r2 * cos(slope), y2 - r2 * sin(slope)
            
            line = plt.Line2D((p1_x, p2_x), (p1_y, p2_y), linewidth=width, color='k')
            
            return plt.gca().add_line(line)
        except:
            print(f"Can't draw line for {SiteObj1.seqNum} & {SiteObj2.seqNum} in {self.molecularName}")
        
        
    @staticmethod
    def mapLinkToSite(Sites, Links):
        siteObj_pair = []
        for link in Links:
            s1, s2 = None, None
            for site in Sites:
                if link.Site1 == site.seqNum:
                    s1 = site
                elif link.Site2 == site.seqNum:
                    s2 = site
            siteObj_pair.append((s1,s2))
        #print(siteObj_pair)
        return siteObj_pair
    
    def displayMolecule(self, Width, saveImage):
        plt.axes()
        coor = np.array([SiteObj.coordinates for SiteObj in self.siteList])
        x_min, x_max = min(coor[:,2]), max(coor[:,2])
        #print(x_min, x_max)
        y_min = min(coor[:,1])
        #print(mod_sites)
        sizes = []
        for site in self.siteList:
            coor = site.coordinates
            x = coor[2]
            x_update = (x - x_min) + 5
            site.coordinates[2] = x_update
            self.draw_circle(site)
            sizes.append(site.radius)
        
        molLength = self.siteList[-1].coordinates[-1] - self.siteList[0].coordinates[-1]
        Largest_ball = max(sizes) + 2
        #print(molLength,  Largest_ball)
        Linked_Sites = self.mapLinkToSite(self.siteList, self.linkList)
        #print(Linked_Sites)
        for s1, s2 in Linked_Sites:
            self.draw_line(s1, s2, width=Width)
        #plt.axes(frameon=False)
        plt.axis('image')
        
        #plt.axis(xmin=0, xmax= (x_max - x_min) + 5, ymin=y_min-Largest_ball, ymax=y_min+Largest_ball)
        
        plt.xlabel('nm', fontsize=16)
        plt.ylabel('nm', fontsize=16)
        plt.tick_params(axis='x', labelsize=16)
        plt.yticks([y_min-Largest_ball,y_min+Largest_ball])
        #plt.xlim((0,35))
        #Axes.set_ylim((0,5))
        
        #plt.tick_params(labelsize='small')
        if saveImage:
            plt.savefig(self.outPath+"\\{}_pyDraw2.png".format(self.molecularName), dpi=400)

        plt.show()
    def displayMolecules(self, sList, linkList, Width=1, saveImage=False):
        y = 2
        for sites, links in zip(sList, linkList):
            
            coor = np.array([SiteObj.coordinates for SiteObj in sites])
            x_min, x_max = min(coor[:,2]), max(coor[:,2])
            sizes = []
            for site in sites:
                coor = site.coordinates
                x = coor[2]
                x_update = (x - x_min) + 1
                site.coordinates[2] = x_update
                site.coordinates[1] = y
                self.draw_circle(site)
                sizes.append(site.radius)
            
            
            Linked_Sites = self.mapLinkToSite(sites, links)
            #print(Linked_Sites)
            for s1, s2 in Linked_Sites:
                self.draw_line(s1, s2, width=Width)
            
            y += 3
           
        plt.axis('image')
        plt.axis(ymin=-1, ymax=7)
        plt.xlabel('nm', fontsize=12)
        plt.ylabel('nm', fontsize=12)
        if saveImage:
            plt.savefig(self.outPath+"/system_pyDraw_test.png", dpi=400)
        plt.show()
                
            
            
     
            
      
class ReadSimFile:
    def __init__(self,txtfile):
        self.txtfile = txtfile
        
    def getOutPath(self):
        path = [f for f in self.txtfile.split('/')]
        newpath = '/'.join(path[:-1])
        outpath = newpath + '/pyStat//pyDraw'
        if not os.path.exists(outpath):
            os.makedirs(outpath)
        return outpath
        
    def readFile(self):
        with open(self.txtfile, 'r') as tmpfile:
            Lines = tmpfile.readlines()
            tmpfile.close()
        return Lines
    
    def getSystemVolume(self):
        Lines = self.readFile()
        Lx, Ly, Lz = 0, 0, 0
        for line in Lines:
            if re.search('L_x', line):
                Lx = float(line.split(':')[-1]) * 1e3  # um to nm
            if re.search('L_y', line):
                Ly = float(line.split(':')[-1]) * 1e3  # um to nm
            if re.search('L_z', line):
                Lz = float(line.split(':')[-1]) * 1e3  # um to nm
        system_vol = Lx*Ly*Lz
        return system_vol
    
    def getSiteStats(self):
        # location: Membrane, extra-cellular or intra-cellular
        # coordinates: [x,y,z]
        Site = namedtuple("Site", ["seqNum", "compartment", "radius", "color", "coordinates"])
        Link = namedtuple("Link", ["Site1", "Site2"])
        LineIndex = []
        molNames, molcounts = [], []
        Lines = self.readFile()
        
        for index, line in enumerate(Lines):
            if not re.search('[*]', line):
                if re.search('MOLECULE', line):
                    specs = line.split(':')[-1].split()
                    if len(specs) > 2:
                        molNames.append(specs[0].replace('"',''))
                        molcounts.append(int(specs[3]))
                        LineIndex.append(index)
        LineIndex.append(len(Lines))
        #print(LineIndex)
        SiteList = []
        LinkList = []
        i = 0
        while i < len(LineIndex) - 1:
            subLines = Lines[LineIndex[i]: LineIndex[i+1]]
            siteList = []
            linkList = []
            for index, line in enumerate(subLines):
                if re.search("SITE", line) and not re.search("[*]", line):
                    line1, line2 = subLines[index+1], subLines[index+2]
                    line = line.split(':')
                    specs = line1.split(':')[-1].split()
                    coors = line2.split()
                    #print(line)
                    siteSpecs = {"seqNum":line[0].strip(), "compartment":line[1].strip(), 
                                 "radius":float(specs[3]), "color":specs[7], 
                                 "coordinates": [float(coors[1]), float(coors[3]), float(coors[5])]}
                    S = Site(**siteSpecs)
                    #print(S)
                    siteList.append(S)
                elif re.search("LINK", line):
                    line = line.split(':')
                    site_pair = {"Site1":line[1].strip().upper(), "Site2":line[4].strip().upper()}
                    linkList.append(Link(**site_pair))
                    
            SiteList.append(siteList)
            LinkList.append(linkList)
            
            i += 1
        
        return molNames, molcounts, SiteList, LinkList
    
    @staticmethod
    def calculateLinkerDistance(linkedSites):
        dist = []
        for site1, site2 in linkedSites:
            p1, p2 = site1.coordinates, site2.coordinates
            d = np.sqrt((p2[0]-p1[0])**2 + (p2[1]-p1[1])**2 + (p2[2]-p1[2])**2)
            dist.append(d)
        return dist
    
    def printStructuralParameters(self, e2eList):
        V = self.getSystemVolume()
        occupied_vol = []
        mols, molcounts, sites, links = self.getSiteStats()
        simName = self.txtfile.split('/')[-1].replace('.txt','')
        print(f'\nSystem: {simName}\n')
        for mol, num, site, link, e2ed in zip(mols, molcounts, sites, links, e2eList):
            linkedSites = Draw_2D_Molecule.mapLinkToSite(site, link)
            LD = self.calculateLinkerDistance(linkedSites)
            #print(LD)
            """
            Caculate the eucledian distance using matrix algebra:
                coor = np.array([s.coordinates for s in site])
                linkerLength = np.sqrt(np.sum((np.diff(coor, axis=0))**2, axis=1))
            """
            
            m_LL = np.mean(LD)
            diameters = np.array([s.radius for s in site])*2
            m_diameter = np.mean(diameters)
            ex_vol = sum(diameters**3)*num
            sp = (e2ed*m_LL)/m_diameter
            occupied_vol.append(ex_vol)
            print(f"molecule: {mol}")
            print(f"mean_LinkerLength: {m_LL:.3f} nm")
            print(f"mean_diameter: {m_diameter:.3f} nm")
            print(f"end_To_end distance: {e2ed:.3f} nm")
            print(f"Excluded volume: {ex_vol:.3f} nm3")
            print(f"SP = {sp:.5f}\n")
        fraction_vol_occupied = (sum(occupied_vol)/V)*100
        print(f"System volume = {V} nm3\nOccupied volume fraction = {fraction_vol_occupied:.3f} %")
            
               
            
    
    
    


import pandas as pd
import matplotlib.pyplot as plt
from SpringSaLaDpy.data_locator import data_file_finder
from SpringSaLaDpy.input_file_extraction import read_input_file
from .Molclustpy_visualization_funcitons import plotAverageZTimeCourse
import statistics
import numpy as np
import os
import csv

def average_Z(df, desired_IDs):
    color_df = df[df['ID'].isin(desired_IDs)]
    return color_df['Z'].astype(float).mean()
'''
def site_info(colors_and_sites):
    print('Sites:')
    i_max = 0
    for i, color_and_site in enumerate(colors_and_sites):
        i_max = i
        append_str = ''
        for site in color_and_site[1]:
            append_str = f'{append_str}{site[1]} of {site[0]}, '
        output = f'{i}: {append_str}'
        print(output[:-2])

    lines = []
    for i in range(0, i_max + 1):
        lines.append(i)

    print('\nList of indices:')
    print(lines)
'''

def site_info(name_list):
    print('Sites:')
    for i, name in enumerate(name_list):
        print(f'{i}: {name[1]} of {name[0]}')
    lines = []
    for i in range(0, len(name_list)):
        lines.append(i)

    print('\nList of indices:')
    print(lines)

def molecule_info(name_list):
    print('Molecules:')
    for i, name in enumerate(name_list):
        print(f'{i}: {name}')
    lines = []
    for i in range(0, len(name_list)):
        lines.append(i)

    print('\nList of indices:')
    print(lines)

def color_info(name_list):
    
    print('Sites with the same color:')
    for i, name in enumerate(name_list):
        print(f'{i}: {name}')
    lines = []
    for i in range(0, len(name_list)):
        lines.append(i)

    print('\nList of indices:')
    print(lines)

def plot(directory_path, mode='mol', indices=[], list_options=True, verbose=False, legend_right=True, fill=True):
    molecules, _ = read_input_file(directory_path)

    new_path = data_file_finder(directory_path, ['viewer_files'], search_term='VIEW_Run0')
    new_directory = os.path.split(new_path)[0]
    num_runs = len(os.listdir(new_directory))

    z_values_list = []
    count = 0
    
    colors_and_sites = []
    used_sites = []

    for molecule in molecules:
        for line in molecule:
            if line[0:4] == 'TYPE':
                color = line.split()[8]
                type = line.split()[2][1:-1]
                molecule_name = molecule[0].split()[1][1:-1]
                if color not in (item[0] for item in colors_and_sites):
                    colors_and_sites.append([color, [[molecule_name, type]]])
                    used_sites.append([molecule_name, type])
                elif [molecule_name, type] not in used_sites:
                    for i, item in enumerate(colors_and_sites):
                        if item[0] == color:
                            colors_and_sites[i][1].append([molecule_name, type])
                    used_sites.append([molecule_name, type])
                else:
                    pass
    
    for run_num in range(num_runs):
        path = data_file_finder(directory_path, ['viewer_files'], run = run_num)
        sites_path = data_file_finder(directory_path, ['data', f'Run{run_num}'], search_term='SiteIDs.csv')

        data_dict = {}

        with open(sites_path, mode='r') as file:
            csv_reader = csv.reader(file)

            for row in csv_reader:
                key = row[0]
                value = row[1].split()
                value = [value[0], value[4]]
                data_dict[key] = value
        file.close()

        with open(path, "r") as file:
            lines = file.readlines()
        file.close()

        molecule_names = []
        site_names = []

        for key in data_dict:
            if data_dict[key][0] not in molecule_names:
                molecule_names.append(data_dict[key][0])

        for key in data_dict:
            if data_dict[key] not in site_names:
                site_names.append(data_dict[key])

        if indices == []:
            if mode=='mol':
                for i in range(len(molecule_names)):
                    indices.append(i)
            elif mode=='site':
                for i in range(len(site_names)):
                    indices.append(i)
            else:
                for i in range(len(colors_and_sites)):
                    indices.append(i)

        if list_options and count == 0:
            if mode == 'mol':
                molecule_info(molecule_names)
            elif mode == 'site':
                site_info(site_names)
            else:
                pass

        count = count + 1
        
        split_file = []
        times = []
        current_list = []

        for line in lines:
            if line == 'SCENE\n':
                split_file.append(current_list)
                current_list = []
            if line[0:2] == 'ID':
                current_list.append(line.strip()[3:])
            if line[0:11] == 'SceneNumber':
                times.append(line.split('\t')[3][:-1])
        split_file.append(current_list)
        _ = split_file.pop(0)

        for i, scene in enumerate(split_file):
            for j, row in enumerate(scene):
                split_file[i][j] = row.split('\t')

        data_frame_list = []

        column_list = ['ID', 'Radius', 'Color', 'X', 'Y', 'Z']

        for scene in split_file:
            data_frame_list.append(pd.DataFrame(scene, columns=column_list))
        
        color_list_full = [x[0] for x in colors_and_sites]
        color_list = []
        if mode=='color':
            if indices == []:
                color_list = color_list_full
            else: 
                for index in indices:
                    color_list.append(colors_and_sites[index][0])
        
        z_values = []
        desired_IDs = []
        legend_list = ['Time']
        if mode == 'mol':    
            for index in indices:
                current_len = len(desired_IDs)
                desired_IDs.append([key for key, value in data_dict.items() if value[0] == molecule_names[index]])
                if len(desired_IDs) != current_len:
                    legend_list.append(molecule_names[index])
        elif mode == 'site':
            for index in indices:
                current_len = len(desired_IDs)
                desired_IDs.append([key for key, value in data_dict.items() if value[0] == site_names[index][0] and value[1] == site_names[index][1]])
                if len(desired_IDs) != current_len:
                    if verbose:
                        legend_list.append(f'Site {site_names[index][1]} of {site_names[index][0]}')
                    else:
                        legend_list.append(site_names[index][1])
        else:
            desired_sites = []
            for color in colors_and_sites:
                tmp=[]
                for site in color[1]:
                    for key, value in data_dict.items():
                        if value == site and site not in tmp:
                            tmp.append(site)
                desired_sites.append(tmp)
            for color in desired_sites:
                tmp=[]
                for site in color:
                    tmp.extend([key for key, value in data_dict.items() if value == site])
                desired_IDs.append(tmp)
            for color in color_list:
                for color_and_site in colors_and_sites:
                    if color == color_and_site[0]:
                        legend_entry = ''
                        item_list = []
                        for site in color_and_site[1]:
                            item_list.append(site)
                        #item_list.sort()
                        for site in item_list:
                            if verbose:
                                legend_entry = legend_entry + f'Site {site[1]} of {site[0]}, '
                            else:
                                legend_entry = legend_entry + f'{site[1]}, ' 
                        legend_list.append(legend_entry[:-2])

        for i in range(len(indices)):
            line = []
            for data_frame in data_frame_list:
                line.append(average_Z(data_frame,desired_IDs[i]))
            z_values.append(line)
        z_values_list.append(z_values)

    if mode == 'color':
        display_colors = []
        for color in color_list_full:
            for color_and_site in colors_and_sites:
                if color == color_and_site[0]:
                    legend_entry = ''
                    item_list = []
                    for site in color_and_site[1]:
                        item_list.append(site)
                    #item_list.sort()
                    for site in item_list:
                        if verbose:
                            legend_entry = legend_entry + f'Site {site[1]} of {site[0]}, '
                        else:
                            legend_entry = legend_entry + f'{site[1]}, ' 
                    display_colors.append(legend_entry[:-2])
        if list_options:
            color_info(display_colors)

    avg_z_values = [[float(time) for time in times]]
    std_z_values = [[float(time) for time in times]]
            
    for i in range(len(indices)):
        tmp_list = []
        for run in z_values_list:
            tmp_list.append(run[i])
        avg = [sum(col)/len(col) for col in zip(*tmp_list)]
        avg_z_values.append(avg)
        if num_runs != 1:
            std = [statistics.stdev(col) for col in zip(*tmp_list)]
            std_z_values.append(std)
        else:
            fill = False
            std = [0 for col in zip(*tmp_list)]
            std_z_values.append(std)
        
    avg_arr = np.transpose(np.array(avg_z_values))
    std_arr = np.transpose(np.array(std_z_values))

    plotAverageZTimeCourse(avg_arr, std_arr, legend_list, legend_right=legend_right, fill=fill, colors=color_list)

    #plt.figure(figsize=(8,5))

    
    


import os
from SpringSaLaDpy.data_locator import find_txt_file
from SpringSaLaDpy.input_file_extraction import read_input_file
from SpringSaLaDpy.time_rounder import find_nearest_time
from .times_2_title import times_2_title

#Switch to the _FOLDER folder if you're not already there and round times to the nearest available values.
def format(search_directory, times, file_type):

    input_file = find_txt_file(search_directory)

    #Round to nearest available time based on dt value
    _, split_file = read_input_file(search_directory)
    if file_type == 'viewer':
        dt = float(split_file[0][5][10:])
    else:
        dt = float(split_file[0][4][9:])
    dt_trimmed =  dt - 1e-12
    total_time = float(split_file[0][1][12:])
    
    total_time = round(total_time - (total_time % dt_trimmed),7)
    rounded_times = []
    for time in times:
        rounded_time = float(find_nearest_time(search_directory, ['data', 'Run0'], time, dt, 'Clusters_Time')[0])
        if rounded_time <= total_time and not rounded_time < 0:
            rounded_times.append(rounded_time)
            if rounded_time != time:
                print(f'Warning, the provided time {time} has been rounded to the closest available value: {rounded_time}')
        else:
            print(f'Warning, the provided time {time} is not between 0 and the last available time: {total_time}')
    if rounded_times == []:
        print(f'Warning, no valid time points. Now defaulting to the last time point: {total_time}')
        rounded_times.append(total_time)

    rounded_times.sort()
    title_str = times_2_title(rounded_times)

    return input_file, rounded_times, title_str



import pandas as pd
import csv

path = r'Examples\Nephrin-Nck-NWasp\Final_version_test_SIMULATIONS\Simulation0_SIM_SIMULATIONS\Simulation0_SIM_FOLDER\data\Run0\SiteIDs.csv'

data_dict = {}

with open(path, mode='r') as file:
    csv_reader = csv.reader(file)

    for row in csv_reader:
        key = row[0]
        value = row[1].split()
        value = [value[0], value[4]]
        data_dict[key] = value

print(data_dict['100000000'])


__all__ = ['Average_z_pos',
           'ClusTopology_ss',
           'DataPy',
           'Histogram',
           'Molclustpy_visualization_funcitons',
           'Plotter',
           'Spatial_Analysis',
           'Cluster_comp',
           'ClusterCrossLinking',
           'Bound_fraction',
           'times_2_title',
           'ssalad_ClusterAnalysis',
           'Composition',
           'Format',
           'Counter',
           'MoleculeCounter']


import pandas as pd
import numpy as np
import os
from SpringSaLaDpy.data_locator import *
from molclustpy import *
from .Molclustpy_visualization_funcitons import *

def format_file_name(file_name):
    split_name = file_name.split('_')
    if 'Site_' not in file_name:
        return split_name[0]
    else:
        return f'{split_name[0]} {split_name[1]} {split_name[2]} {split_name[3]}'

def plot(path, data_selection=[]):
    
    if type(data_selection) != list:
        print("Error: Your data_selection must be a list")
        return
    
    df = pd.read_csv(path, skiprows=1)
    
    output_columns = []
    molecule_list = []
    states = ['free', 'bound', 'total']
    num_entries = (df.shape[1] - 3)/2
    
    #Get list of molecule names from columns
    '''for item in df.columns:
        if item != ' ':
            split_item = item.split().pop()
            if '.1' not in split_item and split_item not in molecule_list and split_item != 'Time':
                molecule_list.append(split_item)'''
    
    #Determine the columns the user wants displayed based on the data_selection argument
    '''for item in data_selection:
        if type(item) == str:
            if item.upper() in (state.upper() for state in states):
                for column in df.columns:
                    if item.upper() in column and not '.1' in column:
                        if column not in output_columns:
                            output_columns.append(column)
            elif item in molecule_list:
                length = len(item)
                for column in df.columns:
                    if column[-length:] == item:
                        if column not in output_columns:
                            output_columns.append(column)
            else:
                print(f'Warning: "{item}" doesn\'t correspond to a recognized molecule name or state in the dataset provided')
        elif type(item) == int:
            if item < 1 or item > num_entries:
                print(f'Warrning: Entry {item} is out of range')
                continue
            if df.columns[item] not in output_columns:
                output_columns.append(df.columns[item])
        else:
            print('Error: Only str and int datatypes are allowed in the data_selection list')
            return'''
    
    output_columns = list(df.columns[1:int(num_entries + 1)])

    outpath = os.path.split(path)[0] + "/pyStat"

    if not os.path.isdir(outpath):
        os.makedirs(outpath)
    
    average_list = ['Time']
    average_list.extend(output_columns)
    stdev_list = [x + '.1' for x in output_columns]
    stdev_list.insert(0,'Time')
    Averages = df[average_list].to_numpy()
    Stdevs = df[stdev_list].to_numpy()
    average_list = [x.replace(' ','_')[1:] if x != 'Time' else x for x in average_list]
    obs_names = '\t'.join(average_list)

    np.savetxt(outpath + "/Mean_Observable_Counts.txt", Averages, header=obs_names, fmt='%.6e')
    np.savetxt(outpath + "/Stdev_Observable_Counts.txt", Stdevs, header=obs_names, fmt='%.6e')
    outpath = os.path.split(outpath)[0]

    file_name = os.path.split(path)[1][23:]
    file_name = format_file_name(file_name)
    plotTimeCourseCopy(outpath, file_name, data_selection)
    

    #Old plotting code
    '''#Plot selected columns with corresponding 1 standard deviations bounds
    if output_columns == []:
        print('Error: No data selected')
        return
    else:
        df.plot('Time', output_columns)
        for output in output_columns:
            plt.fill_between(df['Time'], 
                             df[output] - df[output + '.1'], 
                             df[output] + df[output + '.1'], 
                             alpha=0.2)

        plt.xlabel('Time (Seconds)')
        plt.ylabel('Average Molecule Counts')
        plt.show()'''


# -*- coding: utf-8 -*-
"""
Created on Thu Feb  7 14:19:54 2019

@author: Ani Chattaraj
"""
from SpringSaLaDpy.DrawingTool import Draw_2D_Molecule, ReadSimFile
from data_locator import find_txt_file

def display_molecules(path):
    
    txtfile = find_txt_file(path)

    #txtfile = "Z:/Nwasp_regulation/SpringSaLaD_Models/Newer_NWASP/2nwasp_Arp23_allostericRxn_corr_SIMULATIONS/2nwasp_arp23_allosRxn_corr_SIM.txt"
    #print(file)

    saveImage = False
    simfile = ReadSimFile(txtfile)
    #outpath = simfile.getOutPath()
    outpath = ''
    molNames, molcounts, SiteList, LinkList = simfile.getSiteStats()

    #print(SiteList)
    LineWidth = 0.75

    for mol, site, link in zip(molNames, SiteList, LinkList):
        
        print(mol + ':')
        mol2D = Draw_2D_Molecule(mol,site,link,outpath)
        mol2D.displayMolecule(LineWidth, saveImage)
        #mol2D.displayMolecules(SiteList,LinkList, LineWidth, saveImage)




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import json
from numpy import array

def plotClusterDistCopy(path, times, sizeRange=[], title_str=''):
    # plotting the cluster size distribution (ACO: average cluster occupancy)
    fix, ax = plt.subplots(figsize=(7,4))
    df = pd.read_csv(path + '/pyStat/Cluster_stat/SteadyState_distribution.csv')
    cs, foTM = df['Cluster size'], df['foTM']

    if len(sizeRange) == 0:
        aco = sum(cs*foTM)
        plt.bar(cs, height=foTM, fc='grey',ec='k', label=f'ACO = {aco:.2f}')
        plt.axvline(aco, ls='dashed', lw=1.5, color='k')
        ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True, min_n_ticks=1))
        plt.xlabel('Cluster Size (molecules)')
        plt.ylabel('Fraction of total molecules')
        plt.title(f'Cluster Size Distribution{title_str}')
        plt.legend()
        plt.show()
    else:
        # sizeRange = [1,10,20]
        # clusters : 1-10, 10-20, >20
        idList = [0]
        #xbar = np.arange(1, len(sizeRange)+1, 1)
        xLab = [f'{sizeRange[i]} - {sizeRange[i+1]}' for i in range(len(sizeRange) - 1)]
        xLab.append(f'> {sizeRange[-1]}')
        
        for size in sizeRange[1:]:
            i = 0
            while cs[i] < size:
                i += 1
            if cs[i] == size:
                idList.append(i+1)
            else:
                idList.append(i)
            
        
        foTM_binned = [sum(foTM[idList[i]: idList[i+1]]) for i in range(len(idList)-1)]
        foTM_binned.append(sum(foTM[idList[-1]:]))
        
        try:
            plt.bar(xLab, foTM_binned, color='grey', ec='k')
            ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True, min_n_ticks=1))
            plt.xlabel('Cluster size range (molecules)')
            plt.ylabel('Fraction of total molecules')
            plt.title(f'Binned Cluster Size Distribution{title_str}')
            plt.ylim(0,1)
            plt.show()
        except:
            print('Invalid size range!! Maximal size range might be higher than largest cluster!')

def getColumns(txtfile):
    # name of observables in gdat file
    with open(txtfile,'r') as tf:
        lines = tf.readlines()
    lines[0] = lines[0][8:-1]
    columns = ['Time']
    columns.extend(lines[0].split('\t'))
    return columns

def plotTimeCourseCopy(path, file_name, obsList=[], legend_right=True, fill=True):
    # plotting the observable time course
    txtfile = path + '/pyStat/Count_stat/Mean_Observable_Counts.txt'
    mean_data = np.loadtxt(path + '/pyStat/Count_stat/Mean_Observable_Counts.txt')
    std_data = np.loadtxt(path + '/pyStat/Count_stat/Stdev_Observable_Counts.txt')
    
    _, numVar = mean_data.shape
    colNames = getColumns(txtfile)
    if len(obsList) == 0:
        for i in range(1, numVar):
            x, y, yerr = mean_data[:,0], mean_data[:,int(i)], std_data[:,int(i)]
            plt.plot(x,y, label=f'{colNames[i]}')
            if fill:
                plt.fill_between(x, y-yerr, y+yerr, alpha=0.2)
    else:
        for i in obsList:
            x, y, yerr = mean_data[:,0], mean_data[:,int(i+1)], std_data[:,int(i+1)]
            plt.plot(x,y, label=f'{colNames[i+1]}')
            if fill:
                plt.fill_between(x, y-yerr, y+yerr, alpha=0.2)

    fill_str = ''
    if fill:
        fill_str = ' (bounds of 1 SD)'

    if not legend_right:
        plt.legend()      
    else:
        plt.legend(bbox_to_anchor=(1.02, 1), loc='upper left')
    plt.xlabel('Time (seconds)')
    plt.ylabel('Observable Counts')
    plt.title(f'{file_name}{fill_str}')
    plt.show()

def plotAverageZTimeCourse(mean_data, std_data, colNames, legend_right=True, fill=False, colors=[]):    
    _, numVar = mean_data.shape
    for i in range(1, numVar):
        x, y, yerr = mean_data[:,0], mean_data[:,int(i)], std_data[:,int(i)]
        if colors != []:
            plt.plot(x,y, label=f'{colNames[i]}', color=colors[i-1])
            if fill:
                plt.fill_between(x, y-yerr, y+yerr, alpha=0.2, color=colors[i-1])
            else:
                pass
        else:
            plt.plot(x,y, label=f'{colNames[i]}')
            if fill:
                plt.fill_between(x, y-yerr, y+yerr, alpha=0.2)
            else:
                pass
    
    fill_str = ''
    if fill:
        fill_str = ' (bounds of 1 SD)'

    if not legend_right:
        plt.legend()      
    else:
        plt.legend(bbox_to_anchor=(1.02, 1), loc='upper left')
    plt.xlabel('Time (seconds)')
    plt.ylabel('Distance (nm)')
    plt.title(f'Average Distance to Membrane{fill_str}')
    plt.show()

def plotBarGraph(xdata, yList, yLabels, title='', width=0.1, alpha=0.5, legend_right=False):
    N_entry = len(yList)
    midVarId = N_entry//2
    fix, ax = plt.subplots()
    if N_entry % 2 == 1:
        # odd number 
        plt.bar(xdata, yList[midVarId], width=width, alpha=alpha, label=yLabels[midVarId])
        idx = 1
        for id_lh in range(0, midVarId):
            plt.bar(xdata - 0.15*idx, yList[id_lh], width=width, alpha=alpha, label=yLabels[id_lh])
            idx += 1
        idx = 1
        for id_rh in range(midVarId+1, N_entry):
            plt.bar(xdata + 0.15*idx, yList[id_rh], width=width, alpha=alpha, label=yLabels[id_rh])
            idx += 1
    else:
        # even number 
        shiftIndex = [0.06] + [0.1]*midVarId
        
        idx = 1
        for id_lh in range(0, midVarId):
            plt.bar(xdata - idx*shiftIndex[idx-1], yList[id_lh], width=width, alpha=alpha, label=yLabels[id_lh])
            idx += 1
        
        idx = 1
        for id_rh in range(midVarId, N_entry):
            plt.bar(xdata + idx*shiftIndex[idx-1], yList[id_rh], width=width, alpha=alpha, label=yLabels[id_rh])
            idx += 1
        pass
    
    ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True, min_n_ticks=1))
    if not legend_right:
        plt.legend()      
    else:
        plt.legend(bbox_to_anchor=(1.02, 1), loc='upper left')
    plt.xlabel('Cluster size (molecules)')
    plt.ylabel('Frequency')
    plt.title(title, pad=12)
    plt.show()

def plotClusterCompositionCopy(path, title_str, specialClusters=[], width=0.1, alpha=0.6, legend_right=True):
    df = pd.read_csv(path)
    csList = df['Clusters']
    if len(specialClusters) == 0:
        mols = df.columns[2:]
        freqList = [df[mol] for mol in mols]
        plotBarGraph(csList, freqList, mols, width=width, alpha=alpha, title=f'Cluster Composition{title_str}', legend_right=legend_right)
    else:
        idx = [i for i in range(len(csList)) if csList[i] in specialClusters]
        df2 = df.iloc[idx]
        mols = df.columns[2:]
        freqList = [df2[mol] for mol in mols]
        plotBarGraph(df2['Clusters'], freqList, mols, width=width, alpha=alpha, title=f'Cluster Composition{title_str}', legend_right=legend_right)


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 18 12:26:28 2019

@author: Ani Chattaraj
"""

import re
import os, sys
from decimal import Decimal
import numpy as np
import pandas as pd
from csv import writer
from collections import defaultdict, OrderedDict
from time import time
from glob import glob
import matplotlib.pyplot as plt

#import multiprocessing as mp


def ProgressBar(jobName, progress, length=40):
    '''

    Parameters
    ----------
    jobName : String
        Name to display.
    progress : float
        Fraction of job done.
    length : integer, optional
        Length of the progress bar. The default is 40.

    Returns
    -------
    Displays a progress bar.

    '''
    completionIndex = round(progress*length)
    msg = "\r{} : [{}] {}%".format(jobName, "*"*completionIndex + "-"*(length-completionIndex), round(progress*100))
    if progress >= 1: msg += "\r\n"
    sys.stdout.write(msg)
    sys.stdout.flush()
    

def displayExecutionTime(func):
    """
    This decorator will calculate time needed to execute the "func" function 
    """
    def wrapper(*args, **kwrgs):
        t1 = time()
        func(*args, **kwrgs)
        t2 = time()
        delta = t2 - t1
        if delta < 60:
            print("Execution time : {:.4f} secs".format(delta))
        else:
            t_min, t_sec = int(delta/60), delta%60
            print(f"Execution time : {t_min} mins {t_sec} secs")
    return wrapper


class RunTimeStatistics:
    
    def __init__(self, txtfile, runs=None):
        '''

        Parameters
        ----------
        txtfile : String
            Location of simulation input file.
        runs : int, optional
            Number of runs. The default is None, which reads the number of runs from input file.

        Returns
        -------
        Collect and save execution times across multiple trials.

        '''
        self.simObj = ReadInputFile(txtfile)
        if runs == None:
            numRuns = self.simObj.getNumRuns()
            self.Runs = [int(i) for i in range(numRuns)]
        else:
            self.Runs = runs
        
    def __repr__(self):
        simfile = self.simObj.txtfile.split('/')[-1]
        info = f"Class : {self.__class__.__name__}\nSystem : {simfile}"
        return info
    
    @staticmethod
    def getRunTime(runfile):
        runTime, timeUnit = None, None
        with open(runfile, 'r') as tmpfile:
            line = tmpfile.readlines()[0]
            tps = re.findall(r"[-+]?\d*\.\d+|\d+", line)
            if len(tps) == 1:
                runTime = float(tps[0])
                timeUnit = "sec"
            elif len(tps) == 2:
                runTime = (float(tps[0])*60 + float(tps[-1]))/60
                timeUnit = "min"
            elif len(tps) == 3:
                runTime = (float(tps[0])*3600 + float(tps[1])*60 + float(tps[2]))/3600
                timeUnit = "hour"
            elif len(tps) == 4:
                runTime = (float(tps[0])*24*3600 + float(tps[1])*3600 + float(tps[2])*60 + float(tps[3]))/3600
                timeUnit = "hour"
        return runTime, timeUnit
     
    @displayExecutionTime 
    def getStat(self, fontsize=16):
        print("Getting runtime stats ...")
        seqNo, RT, units = [], [], []
        inpath = self.simObj.getInpath() + "/data"
        outpath = self.simObj.getOutpath("RunTime_stat")
        N_runs = len(self.Runs)
        for i, run in enumerate(self.Runs):
            try:
                rf = inpath + f"/Run{run}/RunningTime.txt"
                rt, unit = self.getRunTime(rf)
                seqNo.append(run)
                RT.append(rt)
                units.append(unit)
                ProgressBar("Progress", (i+1)/N_runs)
            except:
                pass
        #print(units)
        if len(set(units)) == 1: # when all units are of same type (sec/min/hour)
            if units[0] == "sec":
                plt.ylabel("Runtime (secs)", fontsize=fontsize)
            elif units[0] == "min":
                plt.ylabel("Runtime (mins)", fontsize=fontsize)
            elif units[0] == "hour":
                plt.ylabel("Runtime (hours)", fontsize=fontsize)
            plt.xlabel("Run Sequence", fontsize=fontsize)
            plt.scatter(seqNo, RT, color='b')
            with open(outpath + '/runTimeInfo.csv', 'w', newline='') as tmpf:
                wf = writer(tmpf)
                wf.writerows(zip(seqNo, RT))
            simName = self.simObj.txtfile.split('/')[-1].replace(".txt","")
            plt.show()
            plt.title(simName)
            plt.savefig(outpath + "/runTime.png", dpi=100)
            plt.close()
            print("Done plotting the stats ...!")
        else:
            print("Time units are different! Can't plot the runtime stats")
        

class ReadInputFile:
    
    def __init__(self, txtfile):
        self.txtfile = txtfile
    
    def readFile(self):
        with open(self.txtfile, 'r') as tmpfile:
            txtfile = [line for line in tmpfile.readlines()]
        return txtfile
    
    def getInpath(self):
        mypath = os.path.split(self.txtfile)[0]
        return os.path.join(mypath)
    
    def getOutpath(self, statName):
        # statName : Name of the analysis; cluster_stat, e2e_stat etc
        inpath = self.getInpath()
        outpath = inpath + f"/pyStat/{statName}"
        if not os.path.isdir(outpath):
            os.makedirs(outpath)
        return outpath
    
    def getMolecules(self):
        lines = self.readFile()
        molNames, molCounts = [], []
        
        for index, line in enumerate(lines):
            if not re.search('[*]', line):
                if re.search('MOLECULE', line):
                    specs = line.split(':')[-1].split()
                    if len(specs)>2:
                        if int(specs[3]) != 0:  # molecules with non-zero count
                            molNames.append(specs[0].replace('"',''))
                            molCounts.append(int(specs[3]))
                            
        return molNames, molCounts
    
    @staticmethod
    def StringSearch(string, lines):
        value = None
        for line in lines:
            if re.search(string, line):
                value = line.split(':')[-1]
                break
        return value
    
    def getTimeStats(self):
        lines = self.readFile()
        t_tot = self.StringSearch("Total time", lines)
        ts = self.StringSearch("dt", lines)  # ts : time step for simulation
        t_data = self.StringSearch("dt_data", lines)
        t_image = self.StringSearch("dt_image", lines)
        return float(t_tot), float(ts), float(t_data), float(t_image)
    
    def getNumRuns(self):
        numRuns = 0
        lines = self.readFile()
        for line in lines:
            if re.search("\\b" + "Runs" + "\\b", line): # \\b : word boundary; avoids the line 'SimultaneousRuns: 1'
                numRuns = line.split(':')[-1]
                break
        return int(numRuns)



class ClusterAnalysis:

    def __init__(self, simfile, t_total=None, runs=None, withMonomer=True):
        '''

        Parameters
        ----------
        simfile : String
            Location of the simulation input file.
        t_total : Float, optional
            Last timepoint upto which analysis will be carried. The default is None, which reads the total time from input file.
        runs : Interger, optional
            Number of runs to include in the analysis. The default is None, which reads the number of Runs from input file.
        withMonomer : Boolean, optional
            Whether to include monomers (cluser size 1) while computing ACO. The default is True.

        Returns
        -------
        Computes average cluster occupancy of the cluster distributions across multiple runs over all the timepoints.

        '''
        
        self.simfileObj = ReadInputFile(simfile)

        tf, dt, dt_data, dt_image = self.simfileObj.getTimeStats()
        self.dt = dt_data
        if t_total == None:
            self.t_total = tf
        else:
            self.t_total = t_total
            
        self.withMonomer = withMonomer
        
        timepoints = np.arange(0, self.t_total+self.dt, self.dt)
        if timepoints[-1] > self.t_total:
            timepoints = timepoints[:-1]
        self.timePoints = timepoints
        
        numRuns = self.simfileObj.getNumRuns()
        if runs == None:
            self.runs = [run for run in range(numRuns)]
        elif type(runs) is int:
            self.runs = [i for i in range(runs)]
        elif type(runs) is list:
            self.runs = runs
        else:
            print("runs must be an integer or a list of runs")
        

    def __repr__(self):
        simfile = self.simfileObj.txtfile.split('/')[-1]
        info = f"\nClass : {self.__class__.__name__}\nSystem : {simfile}\nTotal_Time : {self.t_total} seconds\t\tnumRuns : {len(self.runs)}\n"
        return info

    @staticmethod
    def getSpeciesCount(df, species):
        try:
            # df : pandas dataframe
            col1, col2 = df[0], df[1]
            return [col2[i] for i in range(len(col1)) if col1[i]== species]
        except:
            pass

    def getClusterAverage(self, csvfile):
        # including the monomer
        acs, aco, foTM = 1, 1, {}
        clusterList = []
        df = pd.read_csv(csvfile, header=None)
        clusterCount = df[1][0]
        #col1, col2 = df[0], df[1]
        if clusterCount == 0:
            print('No cluster found')
        else:
            oligos = self.getSpeciesCount(df, "Size")
            if len(oligos) == 0:
                acs, aco = 1, 1
            else:
                monomerCount = clusterCount - len(oligos)
                clusterList.extend(oligos)
                clusterList.extend([1]*monomerCount)

                N, TM = len(clusterList), sum(clusterList)
                acs = TM/N # acs : average cluster size
                foTM = {clus: (clusterList.count(clus)*(clus/TM)) for clus in set(clusterList)} #foTM : fraction of total molecules
                aco = sum([cs*f for cs, f in foTM.items()])  # aco : average cluster occupancy

        return acs, aco, foTM


    def getOligoAverage(self, csvfile):
        # excluding the monomer
        acs, aco = 0, 0
        df = pd.read_csv(csvfile, header=None)
        oligos = self.getSpeciesCount(df, "Size")
        N, TMC = len(oligos), sum(oligos)  # TMC: total molecules in cluster
        try:
            acs = TMC/N
            foTM = {oligo: (oligos.count(oligo)*(oligo/TMC)) for oligo in set(oligos)}
            aco = sum([cs*f for cs, f in foTM.items()])
        except:
            pass # if there is no cluster > 1, acs = aco = 0
        return acs, aco


    @staticmethod
    def getTimeAverage(ndList, numRuns, numTimePoints):
        ndArray = np.array(ndList).reshape(numRuns, numTimePoints)
        timeAverage = np.mean(ndArray, axis=0, dtype=np.float64) #float64 gives more accurate mean
        return timeAverage

    
    @displayExecutionTime
    def getMeanTrajectory(self, SingleTraj=False):

        simObj = self.simfileObj
        inpath = simObj.getInpath()
        if not self.withMonomer:
            outpath = simObj.getOutpath('Cluster_stat_wom')
        else:
            outpath = simObj.getOutpath("Cluster_stat")

        print('\n' + inpath)

        nf = abs(Decimal(str(self.dt)).as_tuple().exponent) # Number of decimal points to format the clusterTime.csv file

        timepoints = self.timePoints
        numRuns = len(self.runs)
        numTimePoints = len(timepoints)
        
        if SingleTraj is False:
            csvfiles = [inpath + f"/data/Run{run}/Clusters_Time_{tp:.{nf}f}.csv" for run in self.runs for tp in timepoints]
            N = len(csvfiles)

            ave_clus_stat = [*map(lambda x: [self.getClusterAverage(x), ProgressBar('TimeCourse_calc', progress=(csvfiles.index(x)+1)/N)] , csvfiles)]
            ave_stat, _ = list(zip(*ave_clus_stat)) # acs_aco, None
            acs, aco, _ = list(zip(*ave_stat))
            
            print('Done')
    
            mean_acs =np.mean(np.array(acs).reshape((numRuns, numTimePoints)), axis=0)
            mean_aco =np.mean(np.array(aco).reshape((numRuns, numTimePoints)), axis=0)
            std_aco =np.std(np.array(aco).reshape((numRuns, numTimePoints)), axis=0)
    
            tp_ms = timepoints*1e3 # second to millisecond
    
            with open(outpath + "/Clustering_dynamics.csv","w", newline='') as outfile:
                wf = writer(outfile,delimiter=',')  # csv writer
                wf.writerow(['Time (ms)','ACS','ACO','std_ACO'])
                wf.writerows(zip(tp_ms, mean_acs, mean_aco, std_aco))
                outfile.close()
        else:
            print(f'Tracking one trajectory - Run{SingleTraj} ...')
            csvfiles = [inpath + f"/data/Run{SingleTraj}/Clusters_Time_{tp:.{nf}f}.csv" for tp in timepoints]
            N = len(csvfiles)
            ave_clus_stat = [*map(lambda x: [self.getClusterAverage(x), ProgressBar('Cluster_ave_calc', progress=(csvfiles.index(x)+1)/N)] , csvfiles)]
            
            ave_stat, _ = list(zip(*ave_clus_stat)) # acs_aco, None
            acs, aco, foTM = list(zip(*ave_stat))
            #print(foTM)
            #mean_acs =np.mean(np.array(acs))
            #mean_aco =np.mean(np.array(aco))
            
            tp_ms = timepoints*1e3 # second to millisecond
            #CD_timecourse = {k:v for k,v in zip(tp_ms, foTM)}
            #for k, v in CD_timecourse.items():
             #   print(k,v)
            
            with open(outpath + f"/Run{SingleTraj}_distribution_dynamics.csv","w", newline='') as of:
               wf = writer(of, delimiter=',')  # csv writer
               wf.writerow(['Time (ms)','foTM'])
               wf.writerows(zip(tp_ms, foTM))
               of.close()

    
            with open(outpath + f"/Run{SingleTraj}_clustering_dynamics.csv","w", newline='') as outfile:
                wf = writer(outfile,delimiter=',')  # csv writer
                wf.writerow(['Time (ms)','ACS','ACO'])
                wf.writerows(zip(tp_ms, acs, aco))
                outfile.close()


    @staticmethod
    def writeComposition(outpath, compo_dict, molecules):
        d = OrderedDict(sorted(compo_dict.items(), key = lambda x:x[0]))
        with open(outpath + "/Clusters_composition.txt","w") as tmpfile:
            tmpfile.write(f"Cluster Size \t {molecules} : frequency\n\n")
            for k, v in d.items():
                unique_comp = [list(x) for x in set(tuple(x) for x in v)]
                freq = [v.count(comp)/len(v) for comp in unique_comp]
                tmpfile.write(f"  {k}\t\t")
                for cmp, occur in zip(unique_comp, freq):
                    cmp = [str(s) for s in cmp]
                    tmpfile.write(",".join(cmp))
                    tmpfile.write(" : {:.2f}%\t".format(occur*100))
                tmpfile.write("\n\n")

    @staticmethod
    def writeDistribution(outpath, cluster_stat):
        TM_eff, N = sum(cluster_stat), len(cluster_stat)
        unique_clusters = sorted(set(cluster_stat))
        freq = [cluster_stat.count(clus)/N for clus in unique_clusters]
        foTM = [cluster_stat.count(clus)*(clus/TM_eff) for clus in unique_clusters]

        with open(outpath+"/SteadyState_distribution.csv", "w", newline='') as tmpfile:
            wf2 = writer(tmpfile)
            wf2.writerow(['Cluster size','frequency','foTM'])
            wf2.writerows(zip(unique_clusters, freq, foTM))
            tmpfile.close()

    @displayExecutionTime
    def getSteadyStateDistribution(self, SS_timePoints):
        #print("Getting steadystate cluster distribution ...")
        simObj = self.simfileObj
        inpath = simObj.getInpath()
        if self.withMonomer:
            outpath = simObj.getOutpath("Cluster_stat")
        else:
            outpath = simObj.getOutpath("Cluster_stat_wom")

        molecules, counts = simObj.getMolecules()
        nf = abs(Decimal(str(self.dt)).as_tuple().exponent) # Number of decimal points to format the clusterTime.csv file
        composition = defaultdict(list) # composition does not track the monomer identity
        #print("nf : ", nf)
        cluster_stat = []
        numRuns = len(self.runs)
        for j, run in enumerate(self.runs):
            for tp in SS_timePoints:
                df = pd.read_csv((inpath + "/data/Run{}/Clusters_Time_{:.{}f}.csv".format(run, tp, nf)), header=None)
                oligos = self.getSpeciesCount(df, "Size")
                mol_in_cluster = [self.getSpeciesCount(df, mol) for mol in molecules]
                for i, clus in enumerate(oligos):
                    comp = [mol_in_cluster[m][i] for m in range(len(mol_in_cluster))]
                    composition[clus].append(comp)

                if not self.withMonomer:
                    cluster_stat.extend(oligos)
                else:
                    monomerCount = df[1][0] - len(oligos)
                    monomers = [1]*monomerCount
                    cluster_stat.extend(monomers)
                    cluster_stat.extend(oligos)

            ProgressBar("Distribution_calc", (j+1)/numRuns)

        self.writeComposition(outpath, composition, molecules)
        self.writeDistribution(outpath, cluster_stat)

        with open(outpath + "/Sampling_stat.txt","w") as tmpfile:
            ss_tp1000 = [(t*1e3) for t in SS_timePoints]
            tmpfile.write("Number of runs : {}\n\n".format(len(self.runs)))
            tmpfile.write(f"Steady state time points (ms): {ss_tp1000}")



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 28 17:25:41 2020

@author: Ani Chattaraj
"""

import numpy as np
import os, re, sys
from time import time
from csv import writer, reader
#from AllBondPlot import PlotAllBond
import matplotlib.pyplot as plt
import pandas as pd
from glob import glob


class MoleculeCounter:
    
    def __init__ (self, inpath, numRuns=[], dataFile='FullStateCountData.csv'):
        self.inpath = inpath
        if type(numRuns) == list:
            self.runs = numRuns
        if type(numRuns) == int:
            self.runs = [_ for _ in range(numRuns)]
        if len(numRuns) == 0:
            self.runs = 0
        self.dataFile = dataFile
    
    
    def getOutpath(self):
        outpath = self.inpath + "/pyStat/Count_stat"
        if not os.path.isdir(outpath):
            os.makedirs(outpath)
        return outpath
    
    @staticmethod
    def getNumRuns(lines):
        numRuns = 0
        for line in lines:
            if re.search("\\b" + "Runs" + "\\b", line): # \\b : word boundary; avoids the line 'SimultaneousRuns: 1'
                numRuns = line.split(':')[-1]
                break
        return int(numRuns)

    
    def getBoxSize(self):
        txtfile = self.inpath.replace('\\','/').split('/')[-1].replace('_FOLDER','.txt')
        
        with open(self.inpath + "/" + txtfile, 'r') as tf:
            for line in tf.readlines():
                if re.search('L_x', line):
                    Lx = float(line.split(':')[-1].strip())
                if re.search('L_y', line):
                    Ly = float(line.split(':')[-1].strip())
                if re.search('L_z_in', line):
                    Lz = float(line.split(':')[-1].strip())
                
                if re.search("\\b" + "Runs" + "\\b", line): # \\b : word boundary; avoids the line 'SimultaneousRuns: 1'
                    numRuns = line.split(':')[-1]
                    break
            
            
        return (Lx*1e3,Ly*1e3,Lz*1e3), int(numRuns)  # in nm
    
    @staticmethod
    def getMoleculeCount(file, mols):
        mol_stat = []
        df = pd.read_csv(file, header=0)
        for mol in mols:
            mol_stat.append(list(df[mol]))
        
        return np.asarray(mol_stat)
    
    def getTimePoints(self):
        file = None
        for run in self.runs:
            testfile = self.inpath + f"/data/Run{run}/{self.dataFile}"
            #print(testfile)
            if os.path.isfile(testfile):
                file = testfile
                break
            else:
                pass
        else:
            print('No run is complete yet')
        
        if file is None:
            return None
        else:
            df = pd.read_csv(file)
            return df['Time']
            
        
    def getMoleculeStat(self, molecules=[]):
        print('Getting molecular counts...')
        molCount = []
        incmp_runs = []
        
        #tp = self.getTimePoints()
        tp = []
        
        (lx,ly,lz), numRun = self.getBoxSize()
        
        sysVol = lx*ly*lz
        
        if self.runs == 0:
            self.runs = [_ for _ in range(numRun)]
        
        for i, run in enumerate(self.runs):
            file = self.inpath + "/data/Run{}/".format(run) + self.dataFile
            
            df = pd.read_csv(file, header=0)
            tp = df['Time']
            if molecules == []:
                molecules = list(df.columns[1:-1])
            mols = self.getMoleculeCount(file, molecules)
            #print(mols)
            molCount.append(mols)
            
           
            
        if len(incmp_runs) > 0:    
            print('Incomplete Runs: ', incmp_runs)
    
        factor = 6.023 * 1e-7  # converts uM to molecules/nm3
        molConc_arr = np.asarray(molCount)/(factor*sysVol)
        
        
        molCount_arr = np.asarray(molCount)

        _, numMols, _ = molConc_arr.shape # shape : numRuns, numMolecules, numTimepoints
        
        mean_molC = np.mean(molConc_arr, axis=0) # axis = 0 gives average molecular counts over multiple trajectories
        std_molC = np.std(molConc_arr, axis=0)
        
        mean_mols = np.mean(molCount_arr, axis=0) # axis = 0 gives average molecular counts over multiple trajectories
        std_mols = np.std(molCount_arr, axis=0)
        
        df = pd.DataFrame({"Time":[_ for _ in tp]})
        df2 = pd.DataFrame({"Time":[_ for _ in tp]})

        mean_arr = np.concatenate((np.array(tp)[:, np.newaxis], np.transpose(mean_mols)), axis=1)
        std_arr = np.concatenate((np.array(tp)[:, np.newaxis], np.transpose(std_mols)), axis=1)

        columns = ['Time']
        columns.extend(molecules)

        np.savetxt(self.getOutpath() + "\Mean_Observable_Counts.txt", mean_arr, header='\t'.join(columns), fmt='%.6e')
        np.savetxt(self.getOutpath() + "\Stdev_Observable_Counts.txt", std_arr, header='\t'.join(columns), fmt='%.6e')
        
        '''
        for i in range(numMols):
            df[molecules[i] + ' (mean uM)'] = [_ for _ in mean_molC[i]]
            df[molecules[i] + ' (stdev uM)'] = [_ for _ in std_molC[i]]
            
            df2[molecules[i] + ' (mean count)'] = [_ for _ in mean_mols[i]]
            df2[molecules[i] + ' (stdev)'] = [_ for _ in std_mols[i]]
            
            plt.plot(tp*1e3, mean_molC[i], label=molecules[i])
            plt.legend(fontsize=16)
            
        df.to_csv(self.getOutpath() + "/Molecular_Concentration.csv", sep=',')
        df2.to_csv(self.getOutpath() + "/Molecular_Counts.csv", sep=',')
        '''       
        print('Wrote data! Output Directory: ' + self.getOutpath())
        '''
        plt.xlabel('Time (ms)', fontsize=16)
        plt.ylabel('Free concentration ($\mu$M)', fontsize=16)
        plt.show()
        '''
        
 
        
''' 
path = 'C:/Users/chatt/Desktop/pytest/springsalad/test_dataset/A5_B5_flex_3nm_2nm_count_40_SIM_FOLDER' 

mols = ['FREE poly_A', 'FREE poly_B']

mc = MoleculeCounter(path)
mc.getMoleculeStat(mols)
'''
 
      
            


import os

def format_directory(search_directory):
    if os.path.split(search_directory)[1][-7:] == '_FOLDER':
        plotting_path = search_directory
    else:
        plotting_path = os.path.join(search_directory, os.path.split(search_directory)[1][:-12] + '_FOLDER')

    return plotting_path